import os
import sys
from pydoc import locate

import click
import pkg_resources

import jesse.helpers as jh

# Hide the "FutureWarning: pandas.util.testing is deprecated." caused by empyrical
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

# Python version validation.
if jh.python_version() < 3.6:
    print(
        jh.color(
            'Jesse requires Python version above 3.6. Yours is {}'.format(jh.python_version()),
            'red'
        )
    )

# fix directory issue
sys.path.insert(0, os.getcwd())

ls = os.listdir('.')
is_jesse_project = 'strategies' in ls and 'config.py' in ls and 'storage' in ls and 'routes.py' in ls


def validate_cwd():
    """
    make sure we're in a Jesse project
    """
    if not is_jesse_project:
        print(
            jh.color(
                'Current directory is not a Jesse project. You must run commands from the root of a Jesse project.',
                'red'
            )
        )
        os._exit(1)


def inject_local_config():
    """
    injects config from local config file
    """
    local_config = locate('config.config')
    from jesse.config import set_config
    set_config(local_config)


def inject_local_routes():
    """
    injects routes from local routes folder
    """
    local_router = locate('routes')
    from jesse.routes import router

    router.set_routes(local_router.routes)
    router.set_extra_candles(local_router.extra_candles)


# inject local files
if is_jesse_project:
    inject_local_config()
    inject_local_routes()


def register_custom_exception_handler():
    """

    :return:
    """
    import sys
    import threading
    import traceback
    import logging
    from jesse.services import logger as jesse_logger
    import click
    from jesse import exceptions

    log_format = "%(message)s"

    if jh.is_livetrading():
        logging.basicConfig(filename='storage/logs/live-trade.txt', level=logging.INFO, filemode='w', format=log_format)
    elif jh.is_paper_trading():
        logging.basicConfig(filename='storage/logs/paper-trade.txt', level=logging.INFO, filemode='w',
                            format=log_format)
    elif jh.is_collecting_data():
        logging.basicConfig(filename='storage/logs/collect.txt', level=logging.INFO, filemode='w',
                            format=log_format)
    elif jh.is_optimizing():
        logging.basicConfig(filename='storage/logs/optimize.txt', level=logging.INFO, filemode='w',
                            format=log_format)
    else:
        logging.basicConfig(level=logging.INFO)

    # main thread
    def handle_exception(exc_type, exc_value, exc_traceback):
        """

        :param exc_type:
        :param exc_value:
        :param exc_traceback:
        :return:
        """
        if issubclass(exc_type, KeyboardInterrupt):
            sys.excepthook(exc_type, exc_value, exc_traceback)
            return

        # handle Breaking exceptions
        if exc_type in [
            exceptions.ConfigException, exceptions.RouteNotFound, exceptions.InvalidRoutes,
            exceptions.CandleNotFoundInDatabase
        ]:
            click.clear()
            print('=' * 30 + ' EXCEPTION TRACEBACK:')
            traceback.print_tb(exc_traceback, file=sys.stdout)
            print("=" * 73)
            print(
                '\n',
                jh.color('Uncaught Exception:', 'red'),
                jh.color('{}: {}'.format(exc_type.__name__, exc_value), 'yellow')
            )
            return

        # send notifications if it's a live session
        if jh.is_live():
            jesse_logger.error(
                '{}: {}'.format(exc_type.__name__, exc_value)
            )

        if jh.is_live() or jh.is_collecting_data():
            logging.error("Uncaught Exception:", exc_info=(exc_type, exc_value, exc_traceback))
        else:
            print('=' * 30 + ' EXCEPTION TRACEBACK:')
            traceback.print_tb(exc_traceback, file=sys.stdout)
            print("=" * 73)
            print(
                '\n',
                jh.color('Uncaught Exception:', 'red'),
                jh.color('{}: {}'.format(exc_type.__name__, exc_value), 'yellow')
            )

        if jh.is_paper_trading():
            print(
                jh.color(
                    'An uncaught exception was raised. Check the log file at:\n{}'.format(
                        'storage/logs/paper-trade.txt'
                    ),
                    'red'
                )
            )
        elif jh.is_livetrading():
            print(
                jh.color(
                    'An uncaught exception was raised. Check the log file at:\n{}'.format(
                        'storage/logs/live-trade.txt'
                    ),
                    'red'
                )
            )
        elif jh.is_collecting_data():
            print(
                jh.color(
                    'An uncaught exception was raised. Check the log file at:\n{}'.format(
                        'storage/logs/collect.txt'
                    ),
                    'red'
                )
            )

    sys.excepthook = handle_exception

    # other threads
    if jh.python_version() >= 3.8:
        def handle_thread_exception(args):
            """

            :param args:
            :return:
            """
            if args.exc_type == SystemExit:
                return

            # handle Breaking exceptions
            if args.exc_type in [
                exceptions.ConfigException, exceptions.RouteNotFound, exceptions.InvalidRoutes,
                exceptions.CandleNotFoundInDatabase
            ]:
                click.clear()
                print('=' * 30 + ' EXCEPTION TRACEBACK:')
                traceback.print_tb(args.exc_traceback, file=sys.stdout)
                print("=" * 73)
                print(
                    '\n',
                    jh.color('Uncaught Exception:', 'red'),
                    jh.color('{}: {}'.format(args.exc_type.__name__, args.exc_value), 'yellow')
                )
                return

            # send notifications if it's a live session
            if jh.is_live():
                jesse_logger.error(
                    '{}: {}'.format(args.exc_type.__name__, args.exc_value)
                )

            if jh.is_live() or jh.is_collecting_data():
                logging.error("Uncaught Exception:", exc_info=(args.exc_type, args.exc_value, args.exc_traceback))
            else:
                print('=' * 30 + ' EXCEPTION TRACEBACK:')
                traceback.print_tb(args.exc_traceback, file=sys.stdout)
                print("=" * 73)
                print(
                    '\n',
                    jh.color('Uncaught Exception:', 'red'),
                    jh.color('{}: {}'.format(args.exc_type.__name__, args.exc_value), 'yellow')
                )

            if jh.is_paper_trading():
                print(
                    jh.color(
                        'An uncaught exception was raised. Check the log file at:\n{}'.format(
                            'storage/logs/paper-trade.txt'
                        ),
                        'red'
                    )
                )
            elif jh.is_livetrading():
                print(
                    jh.color(
                        'An uncaught exception was raised. Check the log file at:\n{}'.format(
                            'storage/logs/live-trade.txt'
                        ),
                        'red'
                    )
                )
            elif jh.is_collecting_data():
                print(
                    jh.color(
                        'An uncaught exception was raised. Check the log file at:\n{}'.format(
                            'storage/logs/collect.txt'
                        ),
                        'red'
                    )
                )

        threading.excepthook = handle_thread_exception


# create a Click group
@click.group()
@click.version_option(pkg_resources.get_distribution("jesse").version)
def cli():
    """

    """
    pass


@cli.command()
@click.argument('exchange', required=True, type=str)
@click.argument('symbol', required=True, type=str)
@click.argument('start_date', required=True, type=str)
def import_candles(exchange, symbol, start_date):
    """
    imports historical candles from exchange
    """
    validate_cwd()
    from jesse.config import config
    config['app']['trading_mode'] = 'import-candles'

    register_custom_exception_handler()

    from jesse.services import db

    from jesse.modes import import_candles_mode

    import_candles_mode.run(exchange, symbol, start_date)

    db.close_connection()


@cli.command()
@click.argument('start_date', required=True, type=str)
@click.argument('finish_date', required=True, type=str)
@click.option('--debug/--no-debug', default=False,
              help='Displays logging messages instead of the progressbar. Used for debugging your strategy.')
@click.option('--csv/--no-csv', default=False,
              help='Outputs a CSV file of all executed trades on completion.')
@click.option('--json/--no-json', default=False,
              help='Outputs a JSON file of all executed trades on completion.')
@click.option('--fee/--no-fee', default=True, help='You can use "--no-fee" as a quick way to set trading fee to zero.')
@click.option('--chart/--no-chart', default=False,
              help='Generates charts of daily portfolio balance and assets price change. Useful for a visual comparision of your portfolio against the market.')
@click.option('--tradingview/--no-tradingview', default=False,
              help="Generates an output that can be copy-and-pasted into tradingview.com's pine-editor too see the trades in their charts.")
def backtest(start_date, finish_date, debug, csv, json, fee, chart, tradingview):
    """
    backtest mode. Enter in "YYYY-MM-DD" "YYYY-MM-DD"
    """
    validate_cwd()

    from jesse.config import config
    config['app']['trading_mode'] = 'backtest'

    register_custom_exception_handler()

    from jesse.services import db
    from jesse.modes import backtest_mode
    from jesse.services.selectors import get_exchange

    # debug flag
    config['app']['debug_mode'] = debug

    # fee flag
    if not fee:
        for e in config['app']['trading_exchanges']:
            config['env']['exchanges'][e]['fee'] = 0
            get_exchange(e).fee = 0

    backtest_mode.run(start_date, finish_date, chart=chart, tradingview=tradingview, csv=csv, json=json)

    db.close_connection()


@cli.command()
@click.argument('start_date', required=True, type=str)
@click.argument('finish_date', required=True, type=str)
@click.argument('optimal_total', required=True, type=int)
@click.option(
    '--debug/--no-debug', default=False,
    help='Displays detailed logs about the genetics algorithm. Use it if you are interested int he genetics algorithm.'
)
def optimize(start_date, finish_date, optimal_total, debug):
    """
    tunes the hyper-parameters of your strategy
    """
    validate_cwd()
    from jesse.config import config
    config['app']['trading_mode'] = 'optimize'

    register_custom_exception_handler()

    # debug flag
    config['app']['debug_mode'] = debug

    from jesse.modes.optimize_mode import optimize_mode

    optimize_mode(start_date, finish_date, optimal_total)


@cli.command()
@click.argument('name', required=True, type=str)
def make_strategy(name):
    """
    generates a new strategy folder from jesse/strategies/ExampleStrategy
    """
    validate_cwd()
    from jesse.config import config

    config['app']['trading_mode'] = 'make-strategy'

    register_custom_exception_handler()

    from jesse.services import strategy_maker

    strategy_maker.generate(name)


@cli.command()
@click.argument('name', required=True, type=str)
def make_project(name):
    """
    generates a new strategy folder from jesse/strategies/ExampleStrategy
    """
    from jesse.config import config

    config['app']['trading_mode'] = 'make-project'

    register_custom_exception_handler()

    from jesse.services import project_maker

    project_maker.generate(name)


@cli.command()
@click.option('--dna/--no-dna', default=False, help='Translates DNA into parameters. Used in optimize mode only')
def routes(dna):
    """
    lists all routes
    """
    validate_cwd()
    from jesse.config import config

    config['app']['trading_mode'] = 'routes'

    register_custom_exception_handler()

    from jesse.modes import routes_mode

    routes_mode.run(dna)


if 'plugins' in ls:
    @cli.command()
    def collect():
        """
        fetches streamed market data such as tickers, trades, and orderbook from
        the WS connection and stores them into the database for later research.
        """
        validate_cwd()

        # set trading mode
        from jesse.config import config
        config['app']['trading_mode'] = 'collect'

        register_custom_exception_handler()

        from plugins.live.collect_mode import run

        run()


    @cli.command()
    @click.option('--testdrive/--no-testdrive', default=False)
    @click.option('--debug/--no-debug', default=False)
    @click.option('--dev/--no-dev', default=False)
    @click.option('--fee/--no-fee', default=True)
    def live(testdrive, debug, dev, fee):
        """
        trades in real-time on exchange with REAL money
        """
        validate_cwd()

        # set trading mode
        from jesse.config import config
        config['app']['trading_mode'] = 'livetrade'
        config['app']['is_test_driving'] = testdrive

        register_custom_exception_handler()

        # debug flag
        config['app']['debug_mode'] = debug

        from plugins.live import init
        from jesse.services.selectors import get_exchange

        # fee flag
        if not fee:
            for e in config['app']['trading_exchanges']:
                config['env']['exchanges'][e]['fee'] = 0
                get_exchange(e).fee = 0

        # inject live config
        init(config)

        # execute live session
        from plugins.live.live_mode import run
        run(dev)


    @cli.command()
    @click.option('--debug/--no-debug', default=False)
    @click.option('--dev/--no-dev', default=False)
    @click.option('--fee/--no-fee', default=True)
    def paper(debug, dev, fee):
        """
        trades in real-time on exchange with PAPER money
        """
        validate_cwd()

        # set trading mode
        from jesse.config import config
        config['app']['trading_mode'] = 'papertrade'

        register_custom_exception_handler()

        # debug flag
        config['app']['debug_mode'] = debug

        from plugins.live import init
        from jesse.services.selectors import get_exchange

        # fee flag
        if not fee:
            for e in config['app']['trading_exchanges']:
                config['env']['exchanges'][e]['fee'] = 0
                get_exchange(e).fee = 0

        # inject live config
        init(config)

        # execute live session
        from plugins.live.live_mode import run
        run(dev)



from jesse.config import config, reset_config
from jesse.enums import exchanges
from jesse.factories import fake_order
from jesse.store import store


def set_up():
    """

    """
    reset_config()
    config['app']['trading_exchanges'] = [exchanges.SANDBOX, exchanges.BITFINEX]
    config['app']['trading_symbols'] = ['BTCUSD', 'ETHUSD']
    store.reset()


def test_add_new_order():
    set_up()

    o1 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'ETHUSD'})
    o2 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'ETHUSD'})
    store.orders.add_order(o1)
    store.orders.add_order(o2)
    assert store.orders.storage['Sandbox-ETHUSD'] == [o1, o2]


def test_order_state_init():
    set_up()

    assert len(store.orders.storage.keys()) == 4


def test_state_order_count():
    set_up()

    assert store.orders.count(exchanges.SANDBOX, 'BTCUSD') == 0
    store.orders.add_order(fake_order())
    assert store.orders.count(exchanges.SANDBOX, 'BTCUSD') == 1
    store.orders.add_order(fake_order())
    assert store.orders.count(exchanges.SANDBOX, 'BTCUSD') == 2


def test_state_order_get_order_by_id():
    set_up()

    o0 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'ETHUSD'})
    o1 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'ETHUSD'})
    o2 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'ETHUSD'})

    store.orders.get_order_by_id(exchanges.SANDBOX, 'ETHUSD', o2.id)

    # return None if does not exist
    assert store.orders.get_order_by_id(exchanges.SANDBOX, 'ETHUSD',
                                        o0.id) == None

    store.orders.add_order(o1)
    store.orders.add_order(o2)
    assert store.orders.get_order_by_id(exchanges.SANDBOX, 'ETHUSD',
                                        o2.id) == o2
def test_state_order_get_orders():
    set_up()

    o1 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'ETHUSD'})
    o2 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'ETHUSD'})
    store.orders.add_order(o1)
    store.orders.add_order(o2)
    assert store.orders.get_orders(exchanges.SANDBOX,
                                   'ETHUSD') == [o1, o2]





from jesse.models import Route
import jesse.helpers as jh
from jesse import exceptions
import sys


class RouterClass:
    """

    """
    def __init__(self):
        self.routes = []
        self.extra_candles = []
        self.market_data = []

    def set_routes(self, routes):
        """

        :param routes:
        """
        self.routes = []

        for r in routes:
            # validate strategy
            strategy_name = r[3]
            if jh.is_unit_testing():
                exists = jh.file_exists(sys.path[0] + '/jesse/strategies/{}/__init__.py'.format(strategy_name))
            else:
                exists = jh.file_exists('strategies/{}/__init__.py'.format(strategy_name))

            if not exists:
                raise exceptions.InvalidRoutes(
                    'A strategy with the name of "{}" could not be found.'.format(r[3]))

            # validate timeframe
            timeframe = r[2]
            if timeframe not in ['1m', '3m', '5m', '15m', '30m', '1h', '2h', '3h', '4h', '6h', '8h', '1D']:
                raise exceptions.InvalidRoutes(
                    'Timeframe "{}" is invalid. Supported timeframes are 1m, 3m, 5m, 15m, 30m, 1h, 2h, 3h, 4h, 6h, 8h, 1D'.format(
                        timeframe)
                )

            self.routes.append(Route(*r))

    def set_market_data(self, routes):
        """

        :param routes:
        """
        self.market_data = []
        for r in routes:
            self.market_data.append(Route(*r))

    def set_extra_candles(self, extra_candles):
        """

        :param extra_candles:
        """
        self.extra_candles = extra_candles


router: RouterClass = RouterClass()



from typing import Union

import numpy as np
import talib


def sar(candles: np.ndarray, acceleration=0.02, maximum=0.2, sequential=False) -> Union[float, np.ndarray]:
    """
    SAR - Parabolic SAR

    :param candles: np.ndarray
    :param acceleration: float - default: 0.02
    :param maximum: float - default: 0.2
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    res = talib.SAR(candles[:, 3], candles[:, 4], acceleration=acceleration, maximum=maximum)

    return res if sequential else res[-1]



import peewee

import jesse.helpers as jh
from jesse.services.db import db


class Candle(peewee.Model):
    """

    """
    id = peewee.UUIDField(primary_key=True)
    timestamp = peewee.BigIntegerField()
    open = peewee.FloatField()
    close = peewee.FloatField()
    high = peewee.FloatField()
    low = peewee.FloatField()
    volume = peewee.FloatField()
    symbol = peewee.CharField()
    exchange = peewee.CharField()

    # partial candles: 5 * 1m candle = 5m candle while 1m == partial candle
    is_partial = True

    class Meta:
        database = db
        indexes = (
            (('timestamp', 'exchange', 'symbol'), True),
        )

    def __init__(self, attributes=None, **kwargs):
        peewee.Model.__init__(self, attributes=attributes, **kwargs)

        if attributes is None:
            attributes = {}

        for a in attributes:
            setattr(self, a, attributes[a])


if not jh.is_unit_testing():
    # create the table
    Candle.create_table()



import hashlib
import math
import os
import sys
import uuid

import arrow
import click
import numpy as np

CACHED_CONFIG = dict()


def app_mode():
    from jesse.config import config
    return config['app']['trading_mode']


def arrow_to_timestamp(arrow_time):
    return arrow_time.timestamp * 1000


def binary_search(arr: list, item) -> int:
    """
    performs a simple binary search on a sorted list

    :param arr: list
    :param item:

    :return: int
    """
    from bisect import bisect_left

    i = bisect_left(arr, item)
    if i != len(arr) and arr[i] == item:
        return i
    else:
        return -1


def clean_orderbook_list(arr):
    return [[float(i[0]), float(i[1])] for i in arr]


def color(msg_text: str, msg_color: str):
    if not msg_text:
        return ''

    if msg_color == 'black':
        return click.style(msg_text, fg='black')
    if msg_color == 'red':
        return click.style(msg_text, fg='red')
    if msg_color == 'green':
        return click.style(msg_text, fg='green')
    if msg_color == 'yellow':
        return click.style(msg_text, fg='yellow')
    if msg_color == 'blue':
        return click.style(msg_text, fg='blue')
    if msg_color == 'magenta':
        return click.style(msg_text, fg='magenta')
    if msg_color == 'cyan':
        return click.style(msg_text, fg='cyan')
    if msg_color in ['white', 'gray']:
        return click.style(msg_text, fg='white')

    raise ValueError('unsupported color')


def convert_number(old_max, old_min, new_max, new_min, old_value):
    """
    convert a number from one range (ex 40-119) to another
    range (ex 0-30) while keeping the ratio.
    """
    # validation
    if old_value > old_max or old_value < old_min:
        raise ValueError('old_value:{} must be within the range. {}-{}'.format(old_value, old_min, old_max))

    old_range = (old_max - old_min)
    new_range = (new_max - new_min)
    new_value = (((old_value - old_min) * new_range) / old_range) + new_min

    return new_value


def date_diff_in_days(date1, date2):
    if type(date1) is not arrow.arrow.Arrow or type(
            date2) is not arrow.arrow.Arrow:
        raise TypeError('dates must be Arrow instances')

    dif = date2 - date1
    return abs(dif.days)


def date_to_timestamp(date: str) -> int:
    """
    converts date string into timestamp. "2015-08-01" => 1438387200000

    :param date: str
    :return: int
    """
    return arrow_to_timestamp(arrow.get(date, 'YYYY-MM-DD'))


def dna_to_hp(strategy_hp, dna: str):
    hp = {}

    for gene, h in zip(dna, strategy_hp):
        if h['type'] is int:
            decoded_gene = int(
                round(
                    convert_number(119, 40, h['max'], h['min'], ord(gene))
                )
            )
        elif h['type'] is float:
            decoded_gene = convert_number(119, 40, h['max'], h['min'], ord(gene))
        else:
            raise TypeError('Only int and float types are implemented')

        hp[h['name']] = decoded_gene

    return hp


def dump_exception():
    """
    a useful debugging helper
    """
    import traceback
    print(traceback.format_exc())
    terminate_app()


def estimate_average_price(order_qty, order_price, current_qty, current_entry_price):
    """Estimates the new entry price for the position.
    This is used after having a new order and updating the currently holding position.

    Arguments:
        order_qty {float} -- qty of the new order
        order_price {float} -- price of the new order
        current_qty {float} -- current(pre-calculation) qty
        current_entry_price {float} -- current(pre-calculation) entry price

    Returns:
        float -- the new/averaged entry price
    """
    return (abs(order_qty) * order_price + abs(current_qty) *
            current_entry_price) / (abs(order_qty) + abs(current_qty))


def estimate_PNL(qty, entry_price, exit_price, trade_type, trading_fee=0):
    qty = abs(qty)
    profit = qty * (exit_price - entry_price)

    if trade_type == 'short':
        profit *= -1

    fee = trading_fee * qty * (entry_price + exit_price)

    return profit - fee


def estimate_PNL_percentage(qty, entry_price, exit_price, trade_type):
    qty = abs(qty)
    profit = qty * (exit_price - entry_price)

    if trade_type == 'short':
        profit *= -1

    return (profit / (qty * entry_price)) * 100


def file_exists(path: str) -> bool:
    return os.path.isfile(path)


def generate_unique_id():
    return str(uuid.uuid4())


def get_arrow(timestamp):
    return arrow.get(timestamp / 1000)


def get_candle_source(candles: np.ndarray, source_type="close") -> np.ndarray:
    """
     Returns the candles corresponding the selected type.

     :param candles: np.ndarray
     :param source_type: string
     :return: np.ndarray
     """

    if source_type == "close":
        return candles[:, 2]
    elif source_type == "high":
        return candles[:, 3]
    elif source_type == "low":
        return candles[:, 4]
    elif source_type == "open":
        return candles[:, 1]
    elif source_type == "volume":
        return candles[:, 5]
    elif source_type == "hl2":
        return (candles[:, 3] + candles[:, 4]) / 2
    elif source_type == "hlc3":
        return (candles[:, 3] + candles[:, 4] + candles[:, 2]) / 3
    elif source_type == "ohlc4":
        return (candles[:, 1] + candles[:, 3] + candles[:, 4] + candles[:, 2]) / 4
    else:
        raise ValueError('type string not recognised')


def get_config(keys: str, default=None):
    """
    Gets keys as a single string separated with "." and returns value.
    Also accepts a default value so that the app would work even if
    the required config value is missing from config.py file.
    Example: get_config('env.logging.order_submission', True)

    :param keys: str
    :param default: None
    :return:
    """
    if not str:
        raise ValueError('keys string cannot be empty')

    if not keys in CACHED_CONFIG:
        from functools import reduce
        from jesse.config import config
        CACHED_CONFIG[keys] = reduce(lambda d, k: d.get(k, default) if isinstance(d, dict) else default,
                                     keys.split("."), config)

    return CACHED_CONFIG[keys]


def get_strategy_class(strategy_name):
    from pydoc import locate

    if is_unit_testing():
        return locate('jesse.strategies.{}.{}'.format(strategy_name, strategy_name))
    else:
        return locate('strategies.{}.{}'.format(strategy_name, strategy_name))


def insecure_hash(msg: str) -> str:
    return hashlib.md5(msg.encode()).hexdigest()


def insert_list(index: int, item, arr: list):
    """
    helper to insert an item in a Python List without removing the item
    """
    if index == -1:
        return arr + [item]

    return arr[:index] + [item] + arr[index:]


def is_backtesting():
    from jesse.config import config
    return config['app']['trading_mode'] == 'backtest'


def is_collecting_data():
    from jesse.config import config
    return config['app']['trading_mode'] == 'collect'


def is_debuggable(debug_item):
    from jesse.config import config
    return is_debugging() and config['env']['logging'][debug_item]


def is_debugging():
    from jesse.config import config
    return config['app']['debug_mode']


def is_importing_candles():
    from jesse.config import config
    return config['app']['trading_mode'] == 'import-candles'


def is_live():
    return is_livetrading() or is_paper_trading()


def is_livetrading():
    from jesse.config import config
    return config['app']['trading_mode'] == 'livetrade'


def is_optimizing():
    from jesse.config import config
    return config['app']['trading_mode'] == 'optimize'


def is_paper_trading():
    from jesse.config import config
    return config['app']['trading_mode'] == 'papertrade'


def is_test_driving():
    from jesse.config import config
    return config['app']['is_test_driving']


def is_unit_testing():
    return "pytest" in sys.modules


def key(exchange, symbol, timeframe=None):
    if timeframe is None:
        return '{}-{}'.format(exchange, symbol)

    return '{}-{}-{}'.format(exchange, symbol, timeframe)


def max_timeframe(timeframes_list):
    from jesse.enums import timeframes

    if timeframes.DAY_1 in timeframes_list:
        return timeframes.DAY_1
    if timeframes.HOUR_8 in timeframes_list:
        return timeframes.HOUR_8
    if timeframes.HOUR_6 in timeframes_list:
        return timeframes.HOUR_6
    if timeframes.HOUR_4 in timeframes_list:
        return timeframes.HOUR_4
    if timeframes.HOUR_3 in timeframes_list:
        return timeframes.HOUR_3
    if timeframes.HOUR_2 in timeframes_list:
        return timeframes.HOUR_2
    if timeframes.HOUR_1 in timeframes_list:
        return timeframes.HOUR_1
    if timeframes.MINUTE_30 in timeframes_list:
        return timeframes.MINUTE_30
    if timeframes.MINUTE_15 in timeframes_list:
        return timeframes.MINUTE_15
    if timeframes.MINUTE_5 in timeframes_list:
        return timeframes.MINUTE_5
    if timeframes.MINUTE_3 in timeframes_list:
        return timeframes.MINUTE_3

    return timeframes.MINUTE_1


def normalize(x, x_min, x_max):
    """
    Rescaling data to have values between 0 and 1
    """
    x_new = (x - x_min) / (x_max - x_min)
    return x_new


def now():
    if not (is_live() or is_collecting_data() or is_importing_candles()):
        from jesse.store import store
        return store.app.time

    return arrow.utcnow().timestamp * 1000


def np_shift(arr: np.ndarray, num: int, fill_value=0):
    result = np.empty_like(arr)

    if num > 0:
        result[:num] = fill_value
        result[num:] = arr[:-num]
    elif num < 0:
        result[num:] = fill_value
        result[:num] = arr[-num:]
    else:
        result[:] = arr

    return result


def opposite_side(s):
    from jesse.enums import sides

    if s == sides.BUY:
        return sides.SELL
    if s == sides.SELL:
        return sides.BUY
    raise ValueError('unsupported side')


def opposite_type(t):
    from jesse.enums import trade_types

    if t == trade_types.LONG:
        return trade_types.SHORT
    if t == trade_types.SHORT:
        return trade_types.LONG
    raise ValueError('unsupported type')


def orderbook_insertion_index_search(arr, target, ascending=True):
    target = target[0]
    lower = 0
    upper = len(arr)

    if ascending:
        while lower < upper:
            x = lower + (upper - lower) // 2
            val = arr[x][0]
            if target == val:
                return True, x
            elif target > val:
                if lower == x:
                    return False, lower + 1
                lower = x
            elif target < val:
                if lower == x:
                    return False, lower
                upper = x
    else:
        while lower < upper:
            x = lower + (upper - lower) // 2
            val = arr[x][0]
            if target == val:
                return True, x
            elif target < val:
                if lower == x:
                    return False, lower + 1
                lower = x
            elif target > val:
                if lower == x:
                    return False, lower
                upper = x


def orderbook_trim_price(p: float, ascending: bool, unit: float):
    """

    :param p:
    :param ascending:
    :param unit:
    :return:
    """
    if ascending:
        trimmed = np.ceil(p / unit) * unit
        if math.log10(unit) < 0:
            trimmed = round(trimmed, abs(int(math.log10(unit))))
        return p if trimmed == p + unit else trimmed

    trimmed = np.ceil(p / unit) * unit - unit
    if math.log10(unit) < 0:
        trimmed = round(trimmed, abs(int(math.log10(unit))))
    return p if trimmed == p - unit else trimmed


def prepare_qty(qty, side):
    if side.lower() in ('sell', 'short'):
        return -abs(qty)

    if side.lower() in ('buy', 'long'):
        return abs(qty)

    raise TypeError()


def python_version() -> float:
    return float('{}.{}'.format(sys.version_info[0], sys.version_info[1]))


def readable_duration(seconds, granularity=2):
    intervals = (
        ('weeks', 604800),  # 60 * 60 * 24 * 7
        ('days', 86400),  # 60 * 60 * 24
        ('hours', 3600),  # 60 * 60
        ('minutes', 60),
        ('seconds', 1),
    )

    result = []

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip('s')
            result.append("{} {}".format(value, name))
    return ', '.join(result[:granularity])


def relative_to_absolute(path: str) -> str:
    return os.path.abspath(path)


def round_price_for_live_mode(price, roundable_price):
    """
    Rounds price(s) based on exchange requirements

    :param price: float
    :param roundable_price: float | nd.array
    :return: float | nd.array
    """
    n = int(math.log10(price))

    if price < 1:
        price_round_precision = abs(n - 4)
    else:
        price_round_precision = 3 - n
        if price_round_precision < 0:
            price_round_precision = 0

    return np.round(roundable_price, price_round_precision)


def round_qty_for_live_mode(price, roundable_qty):
    """
    Rounds qty(s) based on exchange requirements

    :param price: float
    :param roundable_qty: float | nd.array
    :return: float | nd.array
    """
    n = int(math.log10(price))

    if price < 1:
        qty_round_precision = 0
    else:
        qty_round_precision = n + 1
        if qty_round_precision > 3:
            qty_round_precision = 3

    return np.round(roundable_qty, qty_round_precision)


def secure_hash(msg: str) -> str:
    return hashlib.sha256(msg.encode()).hexdigest()


def should_execute_silently() -> bool:
    return is_optimizing() or is_unit_testing()


def side_to_type(s):
    from jesse.enums import trade_types, sides

    if s == sides.BUY:
        return trade_types.LONG
    if s == sides.SELL:
        return trade_types.SHORT
    raise ValueError


def string_after_character(string: str, character: str):
    try:
        return string.split(character, 1)[1]
    except IndexError:
        return None


def style(msg_text: str, msg_style: str):
    if msg_style is None:
        return msg_text

    if msg_style.lower() in ['bold', 'b']:
        return click.style(msg_text, bold=True)

    if msg_style.lower() in ['underline', 'u']:
        return click.style(msg_text, underline=True)

    raise ValueError('unsupported style')


def terminate_app():
    # close the database
    from jesse.services.db import close_connection
    close_connection()
    # disconnect python from the OS
    os._exit(1)


def timeframe_to_one_minutes(timeframe):
    from jesse.enums import timeframes
    from jesse.exceptions import InvalidTimeframe

    dic = {
        timeframes.MINUTE_1: 1,
        timeframes.MINUTE_3: 3,
        timeframes.MINUTE_5: 5,
        timeframes.MINUTE_15: 15,
        timeframes.MINUTE_30: 30,
        timeframes.HOUR_1: 60,
        timeframes.HOUR_2: 60 * 2,
        timeframes.HOUR_3: 60 * 3,
        timeframes.HOUR_4: 60 * 4,
        timeframes.HOUR_6: 60 * 6,
        timeframes.HOUR_8: 60 * 8,
        timeframes.DAY_1: 60 * 24,
    }

    try:
        return dic[timeframe]
    except KeyError:
        raise InvalidTimeframe(
            'Timeframe "{}" is invalid. Supported timeframes are 1m, 3m, 5m, 15m, 30m, 1h, 2h, 3h, 4h, 6h, 8h, 1D'.format(
                timeframe))


def timestamp_to_arrow(timestamp):
    return get_arrow(timestamp)


def timestamp_to_date(timestamp: int) -> str:
    return str(arrow.get(timestamp / 1000))[:10]


def timestamp_to_time(timestamp):
    return str(arrow.get(timestamp / 1000))


def today() -> int:
    """
    returns today's (beginning) timestamp

    :return: int
    """
    return arrow.utcnow().floor('day').timestamp * 1000


def type_to_side(t):
    from jesse.enums import trade_types, sides

    if t == trade_types.LONG:
        return sides.BUY
    if t == trade_types.SHORT:
        return sides.SELL
    raise ValueError


def unique_list(arr) -> list:
    """
    returns a unique version of the list while keeping its order
    :param arr: list | tuple
    :return: list
    """
    seen = set()
    seen_add = seen.add
    return [x for x in arr if not (x in seen or seen_add(x))]


def floor_with_precision(num, precision=0):
    temp = 10 ** precision
    return math.floor(num * temp) / temp



import jesse.helpers as jh
from jesse.routes import router
from jesse.services import table


def run(dna=False):
    # trading routes
    arr = []
    if not dna:
        print(
            jh.color('{}{}{}'.format('#' * 25, ' Trading Routes ', '#' * 25), 'blue')
        )
        arr.append(('exchange', 'symbol', 'timeframe', 'strategy name', 'DNA'))
    else:
        print(
            jh.color('Translated DNAs into hyper-parameters:', 'blue')
        )

    translated_DNAs_count = 0
    for i, r in enumerate(router.routes):
        if dna and r.dna:
            translated_DNAs_count += 1
            StrategyClass = jh.get_strategy_class(r.strategy_name)
            hyperparameters = jh.dna_to_hp(StrategyClass.hyperparameters(None), r.dna)
            table.key_value(hyperparameters.items(), r.strategy_name, uppercase_title=False)
            print('\n')
        else:
            arr.append([r.exchange, r.symbol, r.timeframe, r.strategy_name, r.dna])
    if not dna:
        table.multi_value(arr)
        print('\n')
    else:
        if not translated_DNAs_count:
            print('No DNA string found.')

    # extra_candles
    if not dna:
        print(
            jh.color('{}{}{}'.format('#' * 25, ' Extra Candles ', '#' * 25), 'blue')
        )
        arr = [('exchange', 'symbol', 'timeframe')]

        for i, r in enumerate(router.extra_candles):
            arr.append([r[0], r[1], r[2]])

        table.multi_value(arr)
        print('\n')



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Make sure to read the docs about routes if you haven't already:
# https://docs.jesse-ai.com/docs/routes.html
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from jesse.utils import anchor_timeframe

# trading routes
routes = [
    ('Binance', 'BTCUSDT', '4h', 'ExampleStrategy'),
]

# in case your strategy requires extra candles, timeframes, ...
extra_candles = [
    ('Binance', 'BTCUSDT', anchor_timeframe('4h')),
]



import requests

import jesse.helpers as jh
from jesse.config import config


def notify(msg):
    """

    :param msg:
    """
    _telegram(msg)


def _telegram(msg):
    token = jh.get_config('env.notifications.telegram_bot_token', '')
    chat_IDs: list = jh.get_config('env.notifications.telegram_chat_IDs', [])

    if not token or not len(chat_IDs) or not config['env']['notifications']['enable_notifications']:
        return

    for id in chat_IDs:
        requests.get(
            'https://api.telegram.org/bot{}/sendMessage?chat_id={}&parse_mode=Markdown&text={}'.format(
                token, id, msg
            )
        )



from collections import namedtuple

import numpy as np
import talib

StochasticFast = namedtuple('StochasticFast', ['k', 'd'])


def stochf(candles: np.ndarray, fastk_period=5, fastd_period=3, fastd_matype=0, sequential=False) -> StochasticFast:
    """
    Stochastic Fast

    :param candles: np.ndarray
    :param fastk_period: int - default=5
    :param fastd_period: int - default=3
    :param fastd_matype: int - default=0
    :param sequential: bool - default=False

    :return: StochasticFast(k, d)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    k, d = talib.STOCHF(
        candles[:, 3],
        candles[:, 4],
        candles[:, 2],
        fastk_period=fastk_period,
        fastd_period=fastd_period,
        fastd_matype=fastd_matype
    )

    if sequential:
        return StochasticFast(k, d)
    else:
        return StochasticFast(k[-1], d[-1])



from collections import namedtuple

import numpy as np
import talib

VWMACD = namedtuple('VWMACD', ['macd', 'signal', 'hist'])


def vwmacd(candles: np.ndarray, fastperiod=12, slowperiod=26, signalperiod=9, sequential=False) -> VWMACD:
    """
    VWMACD - Volume Weighted Moving Average Convergence/Divergence

    :param candles: np.ndarray
    :param fastperiod: int - default: 12
    :param slow_period: int - default: 26
    :param signal_period: int - default: 9
    :param sequential: bool - default: False

    :return: VWMACD(macd, signal, hist)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    vwma_slow = talib.SMA(candles[:, 2] * candles[:, 5], slowperiod) / talib.SMA(candles[:, 5], slowperiod)
    vwma_fast = talib.SMA(candles[:, 2] * candles[:, 5], fastperiod) / talib.SMA(candles[:, 5], fastperiod)
    vwmacd = vwma_fast - vwma_slow
    signal = talib.EMA(vwmacd, signalperiod)
    hist = vwmacd - signal

    if sequential:
        return VWMACD(vwmacd, signal, hist)
    else:
        return VWMACD(vwmacd[-1], signal[-1], hist[-1])



from collections import namedtuple

import numpy as np
import tulipy as ti

FisherTransform = namedtuple('FisherTransform', ['fisher', 'signal'])


def fisher(candles: np.ndarray, period=9, sequential=False) -> FisherTransform:
    """
    The Fisher Transform helps identify price reversals.

    :param candles: np.ndarray
    :param period: int - default: 9
    :param sequential: bool - default=False

    :return: FisherTransform(fisher, signal)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    fisher, fisher_signal = ti.fisher(np.ascontiguousarray(candles[:, 3]), np.ascontiguousarray(candles[:, 4]),
                                      period=period)

    if sequential:
        return FisherTransform(np.concatenate((np.full((candles.shape[0] - fisher.shape[0]), np.nan), fisher), axis=0),
                               np.concatenate(
                                   (np.full((candles.shape[0] - fisher_signal.shape[0]), np.nan), fisher_signal)))
    else:
        return FisherTransform(fisher[-1], fisher_signal[-1])



import empyrical
import numpy as np
import pandas as pd

import jesse.helpers as jh
from jesse.store import store


def candles(candles_array):
    period = jh.date_diff_in_days(jh.get_arrow(candles_array[0][0]), jh.get_arrow(candles_array[-1][0])) + 1

    if period > 365:
        duration = '{} days ({} years)'.format(period, round(period / 365, 2))
    elif period > 30:
        duration = '{} days ({} months)'.format(period, round(period / 30, 2))
    else:
        duration = '{} days'.format(period)

    return [
        ['period', duration],
        ['starting-ending date', '{} => {}'.format(jh.timestamp_to_time(candles_array[0][0])[:10],
                                                   jh.timestamp_to_time(candles_array[-1][0] + 60_000)[:10])],
    ]


def routes(routes):
    array = []

    # header
    array.append(['exchange', 'symbol', 'timeframe', 'strategy', 'DNA'])

    for r in routes:
        array.append([
            r.exchange,
            r.symbol,
            r.timeframe,
            r.strategy_name,
            r.dna
        ])

    return array


def trades(trades_list: list, daily_balance: list):
    starting_balance = 0
    current_balance = 0

    for e in store.exchanges.storage:
        starting_balance += store.exchanges.storage[e].starting_balance
        current_balance += store.exchanges.storage[e].balance

    starting_balance = round(starting_balance, 2)
    current_balance = round(current_balance, 2)

    if len(trades_list) == 0:
        return None

    df = pd.DataFrame.from_records([t.to_dict() for t in trades_list])

    total_completed = len(df)
    winning_trades = df.loc[df['PNL'] > 0]
    total_winning_trades = len(winning_trades)
    losing_trades = df.loc[df['PNL'] < 0]
    total_losing_trades = len(losing_trades)

    losing_i = df['PNL'] < 0
    losing_streaks = losing_i.ne(losing_i.shift()).cumsum()
    losing_streak = losing_streaks[losing_i].value_counts().max()

    winning_i = df['PNL'] > 0
    winning_streaks = winning_i.ne(winning_i.shift()).cumsum()
    winning_streak = winning_streaks[winning_i].value_counts().max()
    largest_losing_trade = round(df['PNL'].min(), 2)
    largest_winning_trade = round(df['PNL'].max(), 2)

    win_rate = len(winning_trades) / (len(losing_trades) + len(winning_trades))
    max_R = round(df['R'].max(), 2)
    min_R = round(df['R'].min(), 2)
    mean_R = round(df['R'].mean(), 2)
    longs_count = len(df.loc[df['type'] == 'long'])
    shorts_count = len(df.loc[df['type'] == 'short'])
    longs_percentage = longs_count / (longs_count + shorts_count) * 100
    short_percentage = 100 - longs_percentage
    fee = df['fee'].sum()
    net_profit = round(df['PNL'].sum(), 2)
    net_profit_percentage = round((net_profit / starting_balance) * 100, 2)
    average_win = round(winning_trades['PNL'].mean(), 2)
    average_loss = round(abs(losing_trades['PNL'].mean()), 2)
    ratio_avg_win_loss = average_win / average_loss
    expectancy = (0 if np.isnan(average_win) else average_win) * win_rate - (
        0 if np.isnan(average_loss) else average_loss) * (1 - win_rate)
    expectancy = round(expectancy, 2)
    expectancy_percentage = round((expectancy / starting_balance) * 100, 2)
    expected_net_profit_every_100_trades = round(expectancy_percentage * 100, 2)
    average_holding_period = df['holding_period'].mean()
    average_winning_holding_period = winning_trades['holding_period'].mean()
    average_losing_holding_period = losing_trades['holding_period'].mean()
    gross_profit = round(df.loc[df['PNL'] > 0]['PNL'].sum(), 2)
    gross_loss = round(df.loc[df['PNL'] < 0]['PNL'].sum(), 2)

    daily_returns = pd.Series(daily_balance).pct_change(1).values
    max_drawdown = round(empyrical.max_drawdown(daily_returns) * 100, 2)
    annual_return = round(empyrical.annual_return(daily_returns) * 100, 2)
    sharpe_ratio = round(empyrical.sharpe_ratio(daily_returns), 2)
    calmar_ratio = round(empyrical.calmar_ratio(daily_returns), 2)
    sortino_ratio = round(empyrical.sortino_ratio(daily_returns), 2)
    omega_ratio = round(empyrical.omega_ratio(daily_returns), 2)
    total_open_trades = store.app.total_open_trades
    open_pl = store.app.total_open_pl

    return {
        'total': np.nan if np.isnan(total_completed) else total_completed,
        'total_winning_trades': np.nan if np.isnan(total_winning_trades) else total_winning_trades,
        'total_losing_trades': np.nan if np.isnan(total_losing_trades) else total_losing_trades,
        'starting_balance': np.nan if np.isnan(starting_balance) else starting_balance,
        'finishing_balance': np.nan if np.isnan(current_balance) else current_balance,
        'win_rate': np.nan if np.isnan(win_rate) else win_rate,
        'max_R': np.nan if np.isnan(max_R) else max_R,
        'min_R': np.nan if np.isnan(min_R) else min_R,
        'mean_R': np.nan if np.isnan(mean_R) else mean_R,
        'ratio_avg_win_loss': np.nan if np.isnan(ratio_avg_win_loss) else ratio_avg_win_loss,
        'longs_count': np.nan if np.isnan(longs_count) else longs_count,
        'longs_percentage': np.nan if np.isnan(longs_percentage) else longs_percentage,
        'short_percentage': np.nan if np.isnan(short_percentage) else short_percentage,
        'shorts_count': np.nan if np.isnan(shorts_count) else shorts_count,
        'fee': np.nan if np.isnan(fee) else fee,
        'net_profit': np.nan if np.isnan(net_profit) else net_profit,
        'net_profit_percentage': np.nan if np.isnan(net_profit_percentage) else net_profit_percentage,
        'average_win': np.nan if np.isnan(average_win) else average_win,
        'average_loss': np.nan if np.isnan(average_loss) else average_loss,
        'expectancy': np.nan if np.isnan(expectancy) else expectancy,
        'expectancy_percentage': np.nan if np.isnan(expectancy_percentage) else expectancy_percentage,
        'expected_net_profit_every_100_trades': np.nan if np.isnan(
            expected_net_profit_every_100_trades) else expected_net_profit_every_100_trades,
        'average_holding_period': average_holding_period,
        'average_winning_holding_period': average_winning_holding_period,
        'average_losing_holding_period': average_losing_holding_period,
        'gross_profit': gross_profit,
        'gross_loss': gross_loss,
        'max_drawdown': max_drawdown,
        'annual_return': annual_return,
        'sharpe_ratio': sharpe_ratio,
        'calmar_ratio': calmar_ratio,
        'sortino_ratio': sortino_ratio,
        'omega_ratio': omega_ratio,
        'total_open_trades': total_open_trades,
        'open_pl': open_pl,
        'winning_streak': winning_streak,
        'losing_streak': losing_streak,
        'largest_losing_trade': largest_losing_trade,
        'largest_winning_trade': largest_winning_trade,
    }



from typing import Union

import numpy as np
import talib


def atr(candles: np.ndarray, period=14, sequential=False) -> Union[float, np.ndarray]:
    """
    ATR - Average True Range

    :param candles: np.ndarray
    :param period: int - default=14
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    res = talib.ATR(candles[:, 3], candles[:, 4], candles[:, 2], timeperiod=period)

    if sequential:
        return res
    else:
        return None if np.isnan(res[-1]) else res[-1]



from typing import Union

import numpy as np
import tulipy as ti



def marketfi(candles: np.ndarray, sequential=False) -> Union[float, np.ndarray]:
    """
    MARKETFI - Market Facilitation Index

    :param candles: np.ndarray
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    res = ti.marketfi(np.ascontiguousarray(candles[:, 3]), np.ascontiguousarray(candles[:, 4]), np.ascontiguousarray(candles[:, 5]))

    return np.concatenate((np.full((candles.shape[0] - res.shape[0]), np.nan), res), axis=0) if sequential else res[-1]



from collections import namedtuple

import numpy as np
import talib

from jesse.helpers import get_candle_source

BollingerBands = namedtuple('BollingerBands', ['upperband', 'middleband', 'lowerband'])


def bollinger_bands(candles: np.ndarray, period=20, devup=2, devdn=2, matype=0, source_type="close",
                    sequential=False) -> BollingerBands:
    """
    BBANDS - Bollinger Bands

    :param candles: np.ndarray
    :param period: int - default: 20
    :param devup: float - default: 2
    :param devdn: float - default: 2
    :param matype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: BollingerBands(upperband, middleband, lowerband)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    upperbands, middlebands, lowerbands = talib.BBANDS(source, timeperiod=period, nbdevup=devup, nbdevdn=devdn,
                                                       matype=matype)

    if sequential:
        return BollingerBands(upperbands, middlebands, lowerbands)
    else:
        return BollingerBands(upperbands[-1], middlebands[-1], lowerbands[-1])



import os
import shutil

import jesse.helpers as jh


def generate(name):
    """

    :param name:
    :return:
    """
    path = '{}'.format(name)

    # validate that doesn't create if current directory is inside a Jesse project
    ls = os.listdir('.')
    is_jesse_project = 'strategies' in ls and 'config.py' in ls and 'storage' in ls and 'routes.py' in ls
    if is_jesse_project:
        print(jh.color('You are already inside a Jesse project. Check your working directory.'.format(name), 'red'))
        return

    # validation for name duplication
    exists = os.path.isdir(path)
    if exists:
        print(jh.color('Project "{}" already exists.'.format(name), 'red'))
        return

    # generate from ExampleStrategy
    dirname, filename = os.path.split(os.path.abspath(__file__))

    shutil.copytree('{}/project_template'.format(dirname), path)

    # output the location of generated strategy directory
    print(jh.color('Your project is created successfully. \nRun "cd {}" to begin algo-trading!'.format(path, name),
                   'green'))



from typing import Union

import numpy as np
import tulipy as ti


def cvi(candles: np.ndarray, period=5, sequential=False) -> Union[float, np.ndarray]:
    """
    CVI - Chaikins Volatility

    :param candles: np.ndarray
    :param period: int - default: 5
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    res = ti.cvi(np.ascontiguousarray(candles[:, 3]), np.ascontiguousarray(candles[:, 4]), period=period)

    return np.concatenate((np.full((candles.shape[0] - res.shape[0]), np.nan), res), axis=0) if sequential else res[-1]



from jesse import exceptions


def validate_routes(router):
    if not router.routes:
        raise exceptions.RouteNotFound(
            'No routes found. Please add at least one route at: routes.py\nMore info: https://docs.jesse-ai.com/docs/routes.html#routing')



from abc import ABC, abstractmethod


class Exchange(ABC):
    """
    The interface that every Exchange driver has to implement
    """

    @abstractmethod
    def market_order(self, symbol, qty, current_price, side, role, flags):
        """

        :param symbol:
        :param qty:
        :param current_price:
        :param side:
        :param role:
        :param flags:
        """
        pass

    @abstractmethod
    def limit_order(self, symbol, qty, price, side, role, flags):
        """

        :param symbol:
        :param qty:
        :param price:
        :param side:
        :param role:
        :param flags:
        """
        pass

    @abstractmethod
    def stop_order(self, symbol, qty, price, side, role, flags):
        """

        :param symbol:
        :param qty:
        :param price:
        :param side:
        :param role:
        :param flags:
        """
        pass

    @abstractmethod
    def cancel_all_orders(self, symbol):
        """

        :param symbol:
        """
        pass

    @abstractmethod
    def cancel_order(self, symbol, order_id):
        """

        :param symbol:
        :param order_id:
        """
        pass

    @abstractmethod
    def get_exec_inst(self, flags):
        """

        :param flags:
        """
        pass



import multiprocessing
import os
import pickle
import sys
from abc import ABC, abstractmethod
from random import randint, choices, choice

# for macOS only
if sys.platform == 'darwin':
    multiprocessing.set_start_method('fork')
from multiprocessing import Process, cpu_count, Manager

import click
import numpy as np
import pydash

import jesse.helpers as jh
from jesse.services import table
from jesse.routes import router
import jesse.services.logger as logger
from jesse.store import store
import jesse.services.report as report
import traceback
import os

class Genetics(ABC):
    def __init__(self, iterations, population_size, solution_len,
                 charset='()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvw',
                 fitness_goal=1,
                 options=None):
        """
        :param iterations: int
        :param population_size: int
        :param solution_len: int
        :param charset: str default= '()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvw'
        which is 40-119 (len=80)
        :param fitness_goal:
        """
        # used for naming the files related to this session
        self.session_id = str(jh.now())
        self.started_index = 0
        self.start_time = jh.now()
        self.population = []
        self.iterations = iterations
        self.population_size = population_size
        self.solution_len = solution_len
        self.charset = charset
        self.fitness_goal = fitness_goal

        if options is None:
            self.options = {}
        else:
            self.options = options

        os.makedirs('./storage/temp/optimize', exist_ok=True)
        self.temp_path = './storage/temp/optimize/{}-{}-{}-{}.pickle'.format(
            self.options['strategy_name'], self.options['exchange'],
            self.options['symbol'], self.options['timeframe']
        )

        if fitness_goal > 1 or fitness_goal < 0:
            raise ValueError('fitness scores must be between 0 and 1')

        # if temp file exists, load data to resume previous session
        if jh.file_exists(self.temp_path):
            if click.confirm('Previous session detected. Do you want to resume?', default=True):
                self.load_progress()

    @abstractmethod
    def fitness(self, dna) -> tuple:
        """
        calculates and returns the fitness score the the DNA

        :param dna: str
        :return: float
        """
        pass

    def generate_initial_population(self):
        """
        generates the initial population
        """
        cores_num = cpu_count()
        loop_length = int(self.population_size / cores_num)

        with click.progressbar(length=loop_length, label='Generating initial population...') as progressbar:
            for i in range(loop_length):
                people = []
                with Manager() as manager:
                    dna_bucket = manager.list([])
                    workers = []

                    def get_fitness(dna, dna_bucket):
                        try:
                            fitness_score, fitness_log = self.fitness(dna)
                            dna_bucket.append((dna, fitness_score, fitness_log))
                        except Exception as e:
                            proc = os.getpid()
                            logger.error('process failed - ID: {}'.format(str(proc)))
                            logger.error("".join(traceback.TracebackException.from_exception(e).format()))
                            raise e

                    try:
                        for _ in range(cores_num):
                            dna = ''.join(choices(self.charset, k=self.solution_len))
                            w = Process(target=get_fitness, args=(dna, dna_bucket))
                            w.start()
                            workers.append(w)

                        # join workers
                        for w in workers:
                            w.join()
                            if w.exitcode > 0:
                                logger.error('a process exited with exitcode: {}'.format(str(w.exitcode)))
                    except KeyboardInterrupt:
                        print(
                            jh.color('Terminating session...', 'red')
                        )

                        # terminate all workers
                        for w in workers:
                            w.terminate()

                        # shutdown the manager process manually since garbage collection cannot won't get to do it for us
                        manager.shutdown()

                        # now we can terminate the main session safely
                        jh.terminate_app()
                    except:
                        raise

                    for d in dna_bucket:
                        people.append({
                            'dna': d[0],
                            'fitness': d[1],
                            'log': d[2]
                        })

                # update dashboard
                click.clear()
                progressbar.update(1)
                print('\n')

                table_items = [
                    ['Started at', jh.get_arrow(self.start_time).humanize()],
                    ['Index', '{}/{}'.format(len(self.population), self.population_size)],
                    ['errors/info', '{}/{}'.format(len(store.logs.errors), len(store.logs.info))],
                    ['Trading Route', '{}, {}, {}, {}'.format(
                        router.routes[0].exchange, router.routes[0].symbol, router.routes[0].timeframe,
                        router.routes[0].strategy_name
                    )],
                    # TODO: add generated DNAs?
                    # ['-'*10, '-'*10],
                    # ['DNA', people[0]['dna']],
                    # ['fitness', round(people[0]['fitness'], 6)],
                    # ['training|testing logs', people[0]['log']],
                ]
                if jh.is_debugging():
                    table_items.insert(3, ['Population Size', self.population_size])
                    table_items.insert(3, ['Iterations', self.iterations])
                    table_items.insert(3, ['Solution Length', self.solution_len])
                    table_items.insert(3, ['-'*10, '-'*10])

                table.key_value(table_items, 'Optimize Mode', alignments=('left', 'right'))

                # errors
                if jh.is_debugging() and len(report.errors()):
                    print('\n')
                    table.key_value(report.errors(), 'Error Logs')

                for p in people:
                    self.population.append(p)

        # sort the population
        self.population = list(sorted(self.population, key=lambda x: x['fitness'], reverse=True))

    def mutate(self, baby):
        replace_at = randint(0, self.solution_len - 1)
        replace_with = choice(self.charset)
        dna = '{}{}{}'.format(baby['dna'][:replace_at], replace_with, baby['dna'][replace_at + 1:])
        fitness_score, fitness_log = self.fitness(dna)

        return {
            'dna': dna,
            'fitness': fitness_score,
            'log': fitness_log
        }

    def make_love(self):
        mommy = self.select_person()
        daddy = self.select_person()

        dna = ''

        for i in range(self.solution_len):
            if i % 2 == 0:
                dna += daddy['dna'][i]
            else:
                dna += mommy['dna'][i]

        fitness_score, fitness_log = self.fitness(dna)

        return {
            'dna': dna,
            'fitness': fitness_score,
            'log': fitness_log
        }

    def select_person(self):
        # len(self.population) instead of self.population_size because some DNAs might not have been created due errors
        random_index = np.random.choice(len(self.population), int(len(self.population) / 100), replace=False)
        chosen_ones = []

        for r in random_index:
            chosen_ones.append(self.population[r])

        return pydash.max_by(chosen_ones, 'fitness')

    def evolve(self):
        """
        the main method, that runs the evolutionary algorithm
        """
        # generate the population if starting
        if self.started_index == 0:
            self.generate_initial_population()
            if len(self.population) < 0.5 * self.population_size:
                raise ValueError('Too many errors: less then half of the planned population size could be generated.')


        cores_num = cpu_count()
        loop_length = int(self.iterations / cores_num)

        i = self.started_index
        with click.progressbar(length=loop_length, label='Evolving...') as progressbar:
            while i < loop_length:
                with Manager() as manager:
                    people = manager.list([])
                    workers = []

                    def get_baby(people):
                        try:
                            # let's make a baby together LOL
                            baby = self.make_love()
                            # let's mutate baby's genes, who knows, maybe we create a x-man or something
                            baby = self.mutate(baby)
                            people.append(baby)
                        except Exception as e:
                            proc = os.getpid()
                            logger.error('process failed - ID: {}'.format(str(proc)))
                            logger.error("".join(traceback.TracebackException.from_exception(e).format()))
                            raise e
                    try:
                        for _ in range(cores_num):
                            w = Process(target=get_baby, args=[people])
                            w.start()
                            workers.append(w)

                        for w in workers:
                            w.join()
                            if w.exitcode > 0:
                                logger.error('a process exited with exitcode: {}'.format(str(w.exitcode)))
                    except KeyboardInterrupt:
                        print(
                            jh.color('Terminating session...', 'red')
                        )

                        # terminate all workers
                        for w in workers:
                            w.terminate()

                        # shutdown the manager process manually since garbage collection cannot won't get to do it for us
                        manager.shutdown()

                        # now we can terminate the main session safely
                        jh.terminate_app()
                    except:
                        raise

                    # update dashboard
                    click.clear()
                    progressbar.update(1)
                    print('\n')

                    table_items = [
                        ['Started At', jh.get_arrow(self.start_time).humanize()],
                        ['Index/Total', '{}/{}'.format((i + 1) * cores_num, self.iterations)],
                        ['errors/info', '{}/{}'.format(len(store.logs.errors), len(store.logs.info))],
                        ['Route', '{}, {}, {}, {}'.format(
                            router.routes[0].exchange, router.routes[0].symbol, router.routes[0].timeframe,
                            router.routes[0].strategy_name
                        )]
                    ]
                    if jh.is_debugging():
                        table_items.insert(
                            3,
                            ['Population Size, Solution Length', '{}, {}'.format(self.population_size, self.solution_len)]
                        )


                    table.key_value(table_items, 'info', alignments=('left', 'right'))

                    # errors
                    if jh.is_debugging() and len(report.errors()):
                        print('\n')
                        table.key_value(report.errors(), 'Error Logs')

                    print('\n')
                    print('Best DNA candidates:')
                    print('\n')

                    # print fittest individuals
                    fittest_list = [['Rank', 'DNA', 'Fitness', 'Training log || Testing log'], ]
                    if self.population_size > 50:
                        number_of_ind_to_show = 15
                    elif self.population_size > 20:
                        number_of_ind_to_show = 10
                    elif self.population_size > 9:
                        number_of_ind_to_show = 9
                    else:
                        raise ValueError('self.population_size cannot be less than 10')

                    for j in range(number_of_ind_to_show):
                        fittest_list.append(
                            [j + 1, self.population[j]['dna'], self.population[j]['fitness'],
                             self.population[j]['log']],
                        )
                    table.multi_value(fittest_list, with_headers=True, alignments=('left', 'left', 'right', 'left'))

                    # one person has to die and be replaced with the newborn baby
                    for baby in people:
                        random_index = randint(0, len(self.population) - 1)
                        try:
                            self.population[random_index] = baby
                        except IndexError:
                            print('=============')
                            print('self.population_size: {}'.format(self.population_size))
                            print('self.population length: {}'.format(len(self.population)))
                            jh.terminate_app()

                        self.population = list(sorted(self.population, key=lambda x: x['fitness'], reverse=True))

                        # reaching the fitness goal could also end the process
                        if baby['fitness'] >= self.fitness_goal:
                            progressbar.update(self.iterations - i)
                            print('\n')
                            print('fitness goal reached after iteration {}'.format(i))
                            return baby

                    # save progress after every n iterations
                    if i != 0 and int(i * cores_num) % 50 == 0:
                        self.save_progress(i)

                    # store a take_snapshot of the fittest individuals of the population
                    if i != 0 and i % int(100 / cores_num) == 0:
                        self.take_snapshot(i * cores_num)

                    i += 1

        print('\n\n')
        print('Finished {} iterations.'.format(self.iterations))

        return self.population

    def run(self):
        return self.evolve()

    def save_progress(self, iterations_index):
        """
        pickles data so we can later resume optimizing
        """
        data = {
            'population': self.population,
            'iterations': self.iterations,
            'iterations_index': iterations_index,
            'population_size': self.population_size,
            'solution_len': self.solution_len,
            'charset': self.charset,
            'fitness_goal': self.fitness_goal,
            'options': self.options,
            'session_id': self.session_id
        }

        with open(self.temp_path, 'wb') as f:
            pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    def load_progress(self):
        """
        unpickles data to resume from previous optimizing session population
        """
        with open(self.temp_path, 'rb') as f:
            data = pickle.load(f)

        self.population = data['population']
        self.iterations = data['iterations']
        self.started_index = data['iterations_index']
        self.population_size = data['population_size']
        self.solution_len = data['solution_len']
        self.charset = data['charset']
        self.fitness_goal = data['fitness_goal']
        self.options = data['options']
        self.session_id = data['session_id']

    def take_snapshot(self, index):
        """
        stores a snapshot of the fittest population members into a file.
        """
        path = './storage/genetics/{}.txt'.format(self.session_id)
        os.makedirs('./storage/genetics', exist_ok=True)
        txt = ''
        with open(path, 'a') as f:
            txt += '\n\n'
            txt += '# iteration {}'.format(index)
            txt += '\n'

            for i in range(30):
                txt += '\n'
                txt += "{} ==  {}  ==  {}  ==  {}".format(
                    i + 1, self.population[i]['dna'], self.population[i]['fitness'], self.population[i]['log']
                )

            f.write(txt)



from .get_candles import get_candles


def init():
    """

    """
    from pydoc import locate
    import os
    import sys
    import jesse.helpers as jh

    # Python version validation.
    if jh.python_version() < 3.6:
        print(
            jh.color(
                'Jesse has not beed tested with your Python version ({}), hence it may not work properly. Consider upgrading to >= 3.7'.format(
                    jh.python_version()),
                'red'
            )
        )

    # fix directory issue
    sys.path.insert(0, os.getcwd())

    ls = os.listdir('.')
    is_jesse_project = 'strategies' in ls and 'config.py' in ls and 'storage' in ls and 'routes.py' in ls

    if not is_jesse_project:
        print(
            jh.color(
                'Invalid directory. To use Jesse inside notebooks, create notebooks inside the root of a Jesse project.',
                'red'
            )
        )

    if is_jesse_project:
        local_config = locate('config.config')
        from jesse.config import set_config
        set_config(local_config)



from typing import Union

import numpy as np
import talib


def ultosc(candles: np.ndarray, timeperiod1=7, timeperiod2=14, timeperiod3=28, sequential=False) -> Union[
    float, np.ndarray]:
    """
    ULTOSC - Ultimate Oscillator

    :param candles: np.ndarray
    :param timeperiod1: int - default=7
    :param timeperiod2: int - default=14
    :param timeperiod3: int - default=28
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    res = talib.ULTOSC(candles[:, 3], candles[:, 4], candles[:, 2], timeperiod1=timeperiod1, timeperiod2=timeperiod2,
                       timeperiod3=timeperiod3)

    if sequential:
        return res
    else:
        return None if np.isnan(res[-1]) else res[-1]



from collections import namedtuple

import numpy as np
import talib
import tulipy as ti

from jesse.helpers import get_candle_source

StochasticRSI = namedtuple('StochasticRSI', ['k', 'd'])


def srsi(candles: np.ndarray, period=14, period_stoch=14, k=3, d=3, source_type="close", sequential=False) -> StochasticRSI:
    """
    Stochastic RSI

    :param candles: np.ndarray
    :param period: int - default: 14 - RSI Length
    :param period_stoch: int - default: 14 - Stochastic Length
    :param k: int - default: 3
    :param d: int - default: 3
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: StochasticRSI(k, d)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    rsi_np = talib.RSI(source, timeperiod=period)
    rsi_np = rsi_np[np.logical_not(np.isnan(rsi_np))]
    fast_k, fast_d = ti.stoch(rsi_np, rsi_np, rsi_np, period_stoch, k, d)

    if sequential:
        fast_k = np.concatenate((np.full((candles.shape[0] - fast_k.shape[0]), np.nan), fast_k), axis=0)
        fast_d = np.concatenate((np.full((candles.shape[0] - fast_d.shape[0]), np.nan), fast_d), axis=0)
        return StochasticRSI(fast_k, fast_d)
    else:
        return StochasticRSI(fast_k[-1], fast_d[-1])



from typing import Union

import numpy as np
import talib


def chande(candles: np.ndarray, period=22, mult=3.0, direction="long", sequential=False) -> Union[float, np.ndarray]:
    """
    Chandelier Exits

    :param candles: np.ndarray
    :param period: int - default: 22
    :param period: float - default: 3.0
    :param direction: str - default: "long"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    candles_close = candles[:, 2]
    candles_high = candles[:, 3]
    candles_low = candles[:, 4]

    atr = talib.ATR(candles_high, candles_low, candles_close, timeperiod=period)

    maxp = np.zeros(len(candles_high) - period + 1)
    if direction == 'long':
        for i in range(period - 1, len(candles_high)):
            maxp[i - period + 1] = np.amax(candles_high[i - period + 1:i + 1])
        maxp = np.concatenate((np.full((candles.shape[0] - maxp.shape[0]), np.nan), maxp))
        result = maxp - atr * mult
    elif direction == 'short':
        for i in range(period - 1, len(candles_high)):
            maxp[i - period + 1] = np.amin(candles_low[i - period + 1:i + 1])
        maxp = np.concatenate((np.full((candles.shape[0] - maxp.shape[0]), np.nan), maxp))
        result = maxp + atr * mult
    else:
        print('The last parameter must be \'short\' or \'long\'')

    return result if sequential else result[-1]



import peewee

import jesse.helpers as jh
from jesse.config import config

if not jh.is_unit_testing():
    # connect to the database
    db = peewee.PostgresqlDatabase(config['env']['databases']['postgres_name'],
                                   user=config['env']['databases']['postgres_username'],
                                   password=config['env']['databases']['postgres_password'],
                                   host=str(config['env']['databases']['postgres_host']),
                                   port=int(config['env']['databases']['postgres_port']))


    def close_connection():
        db.close()


    # connect
    db.connect()
else:
    db = None



import os
import pickle
from time import time


class Cache:
    """

    """
    def __init__(self, path):
        self.path = path

        # make sure path exists
        os.makedirs(path, exist_ok=True)

        # if cache_database exists, load the dictionary
        if os.path.isfile(self.path + "cache_database.pickle"):
            with open(self.path + "cache_database.pickle", 'rb') as f:
                self.db = pickle.load(f)
        # if not, create a dict object. We'll create the file when using set_value()
        else:
            self.db = {}

    def set_value(self, key: str, data, expire_seconds=60 * 60):
        """

        :param key:
        :param data:
        :param expire_seconds:
        """
        # add record into the database
        expire_at = None if expire_seconds is None else time() + expire_seconds
        data_path = self.path + "{}.pickle".format(key)
        self.db[key] = {
            'expire_seconds': expire_seconds,
            'expire_at': expire_at,
            'path': data_path,
        }
        self._update_db()

        # store file
        with open(data_path, 'wb') as f:
            pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    def get_value(self, key: str):
        """

        :param key:
        :return:
        """
        try:
            item = self.db[key]
        except KeyError:
            return False

        # if expired, remove file, and database record
        if item['expire_at'] is not None and time() > item['expire_at']:
            os.remove(item['path'])
            del self.db[key]
            self._update_db()
            return False

        # renew cache expiration time
        if item['expire_at'] is not None:
            item['expire_at'] = time() + item['expire_seconds']
            self._update_db()

        with open(item['path'], 'rb') as f:
            return pickle.load(f)

    def _update_db(self):
        # store/update database
        with open(self.path + "cache_database.pickle", 'wb') as f:
            pickle.dump(self.db, f, protocol=pickle.HIGHEST_PROTOCOL)

    def flush(self):
        for key, item in self.db.items():
            os.remove(item['path'])
        self.db = {}


cache = Cache("storage/temp/")



import jesse.helpers as jh
from jesse.services.notifier import notify


def info(msg):
    """

    :param msg:
    """
    from jesse.store import store

    store.logs.info.append({'time': jh.now(), 'message': msg})

    if (jh.is_backtesting() and jh.is_debugging()) or jh.is_collecting_data():
        print(jh.color('[{}]: {}'.format(jh.timestamp_to_time(jh.now()), msg), 'magenta'))

    if jh.is_live():
        msg = '[INFO | {}] '.format(jh.timestamp_to_time(jh.now())[:19]) + str(msg)
        import logging
        logging.info(msg)


def error(msg):
    """

    :param msg:
    """
    from jesse.store import store

    if jh.is_live() and jh.get_config('env.notifications.events.errors', True):
        notify('ERROR:\n{}'.format(msg))
    if (jh.is_backtesting() and jh.is_debugging()) or jh.is_collecting_data():
        print(jh.color('[{}]: {}'.format(jh.timestamp_to_time(jh.now()), msg), 'red'))

    store.logs.errors.append({'time': jh.now(), 'message': msg})

    if jh.is_live() or jh.is_optimizing():
        msg = '[ERROR | {}] '.format(jh.timestamp_to_time(jh.now())[:19]) + str(msg)
        import logging
        logging.error(msg)



import requests

import jesse.helpers as jh
from jesse import exceptions
from .interface import CandleExchange


class Coinbase(CandleExchange):
    """

    """
    def __init__(self):
        super().__init__('Coinbase', 300, 0.6)
        self.endpoint = 'https://api.pro.coinbase.com/products'

    def init_backup_exchange(self):
        from .bitfinex import Bitfinex
        self.backup_exchange = Bitfinex()

    def get_starting_time(self, symbol: str):
        """
        Because Coinbase's API sucks and does not make this take easy for us,
        we do it manually for as much symbol as we can!

        :param symbol: str
        :return: int
        """
        if symbol == 'BTCUSD':
            return 1438387200000
        elif symbol == 'ETHUSD':
            return 1464739200000
        elif symbol == 'LTCUSD':
            return 1477958400000

        return None

    def fetch(self, symbol: str, start_timestamp: int):
        """
        note1: unlike Bitfinex, Binance does NOT skip candles with volume=0.
        note2: like Bitfinex, start_time includes the candle and so does the end_time.
        """
        end_timestamp = start_timestamp + (self.count - 1) * 60000

        payload = {
            'granularity': '60',
            'start': jh.timestamp_to_time(start_timestamp),
            'end': jh.timestamp_to_time(end_timestamp),
        }

        dashed_symbol = symbol[:3] + '-' + symbol[3:]
        response = requests.get(
            self.endpoint + '/{}/candles'.format(dashed_symbol),
            params=payload
        )

        self._handle_errors(response)

        data = response.json()
        candles = []

        for d in data:
            candles.append({
                'id': jh.generate_unique_id(),
                'symbol': symbol,
                'exchange': self.name,
                'timestamp': int(d[0]) * 1000,
                'open': float(d[3]),
                'close': float(d[4]),
                'high': float(d[2]),
                'low': float(d[1]),
                'volume': float(d[5])
            })

        return candles

    @staticmethod
    def _handle_errors(response):
        # Exchange In Maintenance
        if response.status_code == 502:
            raise exceptions.ExchangeInMaintenance('ERROR: 502 Bad Gateway. Please try again later')

        # unsupported symbol
        if response.status_code == 404:
            raise ValueError(response.json()['message'])

        if response.status_code != 200:
            raise Exception(response.content)



import jesse.helpers as jh
import jesse.services.logger as logger
import jesse.services.selectors as selectors
from jesse.config import config
from jesse.enums import order_statuses, order_flags
from jesse.services.notifier import notify


class Order():
    """

    """
    def __init__(self, attributes=None):
        # id generated by Jesse for database usage
        self.id = ''
        # id generated by market
        self.exchange_id = ''
        self.symbol = ''
        self.exchange = ''
        self.side = ''
        self.type = ''
        self.flag = ''
        self.qty = 0
        self.price = 0
        self.status = order_statuses.ACTIVE
        self.created_at = None
        self.executed_at = None
        self.canceled_at = None
        self.role = None

        if attributes is None:
            attributes = {}

        for a in attributes:
            setattr(self, a, attributes[a])

        if self.created_at is None:
            self.created_at = jh.now()

        p = selectors.get_position(self.exchange, self.symbol)
        if p:
            p._on_opened_order(self)

        if jh.is_live() and config['env']['notifications']['events']['submitted_orders']:
            self.notify_submission()

        if jh.is_debuggable('order_submission'):
            logger.info(
                '{} order: {}, {}, {}, {}, ${}'.format(
                    'QUEUED' if self.is_queued else 'SUBMITTED',
                    self.symbol, self.type, self.side, self.qty,
                    round(self.price, 2)
                )
            )

    def notify_submission(self):
        notify(
            '{} order: {}, {}, {}, {}, ${}'.format(
                'QUEUED' if self.is_queued else 'SUBMITTED',
                self.symbol, self.type, self.side, self.qty,
                round(self.price, 2)
            )
        )

    @property
    def is_canceled(self) -> bool:
        return self.status == order_statuses.CANCELED

    @property
    def is_active(self) -> bool:
        return self.status == order_statuses.ACTIVE

    @property
    def is_queued(self) -> bool:
        """
        Used in live mode only: it means the strategy has considered the order as submitted,
        but the exchange does not accept it because of the distance between the current
        price and price of the order. Hence it's been queued for later submission.

        :return: bool
        """
        return self.status == order_statuses.QUEUED

    @property
    def is_new(self) -> bool:
        return self.is_active

    @property
    def is_executed(self) -> bool:
        return self.status == order_statuses.EXECUTED

    @property
    def is_filled(self) -> bool:
        return self.is_executed

    @property
    def is_reduce_only(self) -> bool:
        return self.flag == order_flags.REDUCE_ONLY

    @property
    def is_close(self) -> bool:
        return self.flag == order_flags.CLOSE

    def cancel(self):
        if self.is_canceled or self.is_executed:
            return

        self.canceled_at = jh.now()
        self.status = order_statuses.CANCELED

        if jh.is_debuggable('order_cancellation'):
            logger.info(
                'CANCELED order: {}, {}, {}, {}, ${}'.format(
                    self.symbol, self.type, self.side, self.qty, round(self.price, 2)
                )
            )

        # notify
        if jh.is_live() and config['env']['notifications']['events']['cancelled_orders']:
            notify(
                'CANCELED order: {}, {}, {}, {}, {}'.format(
                    self.symbol, self.type, self.side, self.qty, round(self.price, 2)
                )
            )

        p = selectors.get_position(self.exchange, self.symbol)
        if p:
            p._on_canceled_order(self)

    def execute(self):
        if self.is_canceled or self.is_executed:
            return

        self.executed_at = jh.now()
        self.status = order_statuses.EXECUTED

        # log
        if jh.is_debuggable('order_execution'):
            logger.info(
                'EXECUTED order: {}, {}, {}, {}, ${}'.format(
                    self.symbol, self.type, self.side, self.qty, round(self.price, 2)
                )
            )

        # notify
        if jh.is_live() and config['env']['notifications']['events']['executed_orders']:
            notify(
                'EXECUTED order: {}, {}, {}, {}, {}'.format(
                    self.symbol, self.type, self.side, self.qty, round(self.price, 2)
                )
            )

        p = selectors.get_position(self.exchange, self.symbol)

        if p:
            p._on_executed_order(self)



from typing import Union

import numpy as np
import talib

from jesse.helpers import get_candle_source


def efi(candles: np.ndarray, period=13, source_type="close", sequential=False) -> Union[float, np.ndarray]:
    """
    EFI - Elders Force Index

    :param candles: np.ndarray
    :param period: int - default: 13
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)

    dif = np.zeros(len(source) - 1)
    for i in range(1, len(source)):
        dif[i - 1] = (source[i] - source[i - 1]) * candles[:, 5][i]

    res = talib.EMA(dif, timeperiod=period)
    res_with_nan = np.concatenate((np.full((candles.shape[0] - res.shape[0]), np.nan), res))

    return res_with_nan if sequential else res_with_nan[-1]



from typing import Union

import numpy as np
import talib

from jesse.helpers import get_candle_source


def t3(candles: np.ndarray, period=5, vfactor=0, source_type="close", sequential=False) -> Union[float, np.ndarray]:
    """
    T3 - Triple Exponential Moving Average (T3)

    :param candles: np.ndarray
    :param period: int - default: 5
    :param vfactor: float - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    res = talib.T3(source, timeperiod=period, vfactor=vfactor)

    return res if sequential else res[-1]



import requests

import jesse.helpers as jh
from jesse import exceptions
from .interface import CandleExchange


class TestnetBinanceFutures(CandleExchange):
    """

    """
    def __init__(self):
        super().__init__('Testnet Binance Futures', 1000, 0.5)
        self.endpoint = 'https://testnet.binancefuture.com/fapi/v1/klines'

    def init_backup_exchange(self):
        from .binance import Binance
        self.backup_exchange = Binance()

    def get_starting_time(self, symbol):
        """

        :param symbol:
        :return:
        """
        payload = {
            'interval': '1d',
            'symbol': symbol,
            'limit': 1500,
        }

        response = requests.get(self.endpoint, params=payload)

        # Exchange In Maintenance
        if response.status_code == 502:
            raise exceptions.ExchangeInMaintenance('ERROR: 502 Bad Gateway. Please try again later')

        # unsupported symbol
        if response.status_code == 400:
            raise ValueError(response.json()['msg'])

        if response.status_code != 200:
            raise Exception(response.content)

        data = response.json()
        first_timestamp = int(data[0][0])
        second_timestamp = first_timestamp + 60_000 * 1440

        return second_timestamp

    def fetch(self, symbol, start_timestamp):
        """

        :param symbol:
        :param start_timestamp:
        :return:
        """
        end_timestamp = start_timestamp + (self.count - 1) * 60000

        payload = {
            'interval': '1m',
            'symbol': symbol,
            'startTime': start_timestamp,
            'endTime': end_timestamp,
            'limit': self.count,
        }

        response = requests.get(self.endpoint, params=payload)

        # Exchange In Maintenance
        if response.status_code == 502:
            raise exceptions.ExchangeInMaintenance('ERROR: 502 Bad Gateway. Please try again later')

        # unsupported symbol
        if response.status_code == 400:
            raise ValueError(response.json()['msg'])

        if response.status_code != 200:
            return

        data = response.json()
        candles = []

        for d in data:
            candles.append({
                'id': jh.generate_unique_id(),
                'symbol': symbol,
                'exchange': self.name,
                'timestamp': int(d[0]),
                'open': float(d[1]),
                'close': float(d[4]),
                'high': float(d[2]),
                'low': float(d[3]),
                'volume': float(d[5])
            })

        return candles



import numpy as np

import jesse.helpers as jh
import jesse.services.selectors as selectors
from jesse.config import config
from jesse.enums import timeframes
from jesse.exceptions import RouteNotFound
from jesse.libs import DynamicNumpyArray
from jesse.models import store_candle_into_db
from jesse.services.candle import generate_candle_from_one_minutes


class CandlesState:
    def __init__(self):
        self.storage = {}
        self.is_initiated = False

    def get_storage(self, exchange, symbol, timeframe):
        key = jh.key(exchange, symbol, timeframe)

        try:
            return self.storage[key]
        except KeyError:
            raise RouteNotFound(
                "Bellow route is required but missing in your routes:\n('{}', '{}', '{}')".format(
                    exchange, symbol, timeframe
                )
            )

    def init_storage(self, bucket_size=1000):
        for exchange in config['app']['considering_exchanges']:
            for symbol in config['app']['considering_symbols']:
                # initiate the '1m' timeframes
                key = jh.key(exchange, symbol, timeframes.MINUTE_1)
                self.storage[key] = DynamicNumpyArray((bucket_size, 6))

                for timeframe in config['app']['considering_timeframes']:
                    key = jh.key(exchange, symbol, timeframe)
                    # ex: 1440 / 60 + 1 (reserve one for forming candle)
                    total_bigger_timeframe = int((bucket_size / jh.timeframe_to_one_minutes(timeframe)) + 1)
                    self.storage[key] = DynamicNumpyArray((total_bigger_timeframe, 6))

    def add_candle(
            self,
            candle: np.ndarray,
            exchange: str,
            symbol: str,
            timeframe: str,
            with_execution=True,
            with_generation=True,
            is_forming_candle=False
    ):
        if jh.is_collecting_data():
            # make sure it's a complete (and not a forming) candle
            if jh.now() >= (candle[0] + 60000):
                store_candle_into_db(exchange, symbol, candle)
            return

        arr: DynamicNumpyArray = self.get_storage(exchange, symbol, timeframe)

        if jh.is_live():
            self.update_position(exchange, symbol, candle)

        # initial
        if len(arr) == 0:
            arr.append(candle)

        # if it's new, add
        elif candle[0] > arr[-1][0]:
            # in paper mode, check to see if the new candle causes any active orders to be executed
            if with_execution and jh.is_paper_trading():
                self.simulate_order_execution(exchange, symbol, timeframe, candle)

            arr.append(candle)

            # generate other timeframes
            if with_generation and timeframe == '1m':
                self.generate_bigger_timeframes(candle, exchange, symbol, with_execution, is_forming_candle)

        # if it's the last candle again, update
        elif candle[0] == arr[-1][0]:
            # in paper mode, check to see if the new candle causes any active orders to get executed
            if with_execution and jh.is_paper_trading():
                self.simulate_order_execution(exchange, symbol, timeframe, candle)

            arr[-1] = candle

            # regenerate other timeframes
            if with_generation and timeframe == '1m':
                self.generate_bigger_timeframes(candle, exchange, symbol, with_execution, is_forming_candle)

        # past candles will be ignored (dropped)
        elif candle[0] < arr[-1][0]:
            return

    @staticmethod
    def update_position(exchange: str, symbol: str, candle: np.ndarray):
        # get position object
        p = selectors.get_position(exchange, symbol)

        # for extra_route candles, p == None, hence no further action is required
        if p is None:
            return

        # update position.current_price
        p.current_price = jh.round_price_for_live_mode(candle[2], candle[2])

    def generate_bigger_timeframes(self, candle: np.ndarray, exchange: str, symbol: str, with_execution: bool,
                                   is_forming_candle: bool):
        if not jh.is_live():
            return

        for timeframe in config['app']['considering_timeframes']:
            # skip '1m'
            if timeframe == '1m':
                continue

            last_candle = self.get_current_candle(exchange, symbol, timeframe)
            generate_from_count = int((candle[0] - last_candle[0]) / 60_000)
            required_for_complete_candle = jh.timeframe_to_one_minutes(timeframe)
            short_candles = self.get_candles(exchange, symbol, '1m')[-1 - generate_from_count:]
            if generate_from_count == (required_for_complete_candle - 1) and not is_forming_candle:
                is_forming_candle = False
            else:
                is_forming_candle = True

            # update latest candle
            generated_candle = generate_candle_from_one_minutes(
                timeframe,
                short_candles,
                True
            )

            self.add_candle(generated_candle, exchange, symbol, timeframe, with_execution, with_generation=False,
                            is_forming_candle=is_forming_candle)

    def simulate_order_execution(self, exchange, symbol, timeframe, new_candle):
        previous_candle = self.get_current_candle(exchange, symbol, timeframe)
        orders = selectors.get_orders(exchange, symbol)

        if previous_candle[2] == new_candle[2]:
            return

        for o in orders:
            # skip inactive orders
            if not o.is_active:
                continue

            if ((o.price >= previous_candle[2]) and (o.price <= new_candle[2])) or (
                    (o.price <= previous_candle[2]) and (o.price >= new_candle[2])):
                o.execute()

    def batch_add_candle(self, candles, exchange, symbol, timeframe, with_generation=True):
        for c in candles:
            self.add_candle(c, exchange, symbol, timeframe, with_execution=False, with_generation=with_generation)

    def forming_estimation(self, exchange, symbol, timeframe):
        long_key = jh.key(exchange, symbol, timeframe)
        short_key = jh.key(exchange, symbol, '1m')
        required_1m_to_complete_count = jh.timeframe_to_one_minutes(timeframe)
        current_1m_count = len(self.get_storage(exchange, symbol, '1m'))

        dif = current_1m_count % required_1m_to_complete_count
        return dif, long_key, short_key

    # # # # # # # # #
    # # # # # getters
    # # # # # # # # #
    def get_candles(self, exchange, symbol, timeframe) -> np.ndarray:
        # no need to worry for forming candles when timeframe == 1m
        if timeframe == '1m':
            arr: DynamicNumpyArray = self.get_storage(exchange, symbol, '1m')
            if len(arr) == 0:
                return np.zeros((0, 6))
            else:
                return arr[:]

        # other timeframes
        dif, long_key, short_key = self.forming_estimation(exchange, symbol, timeframe)
        long_count = len(self.get_storage(exchange, symbol, timeframe))
        short_count = len(self.get_storage(exchange, symbol, '1m'))

        if dif == 0 and long_count == 0:
            return np.zeros((0, 6))

        # complete candle
        if dif == 0 or self.storage[long_key][:long_count][-1][0] == self.storage[short_key][short_count - dif][0]:
            return self.storage[long_key][:long_count]
        # generate forming
        else:
            return np.concatenate(
                (
                    self.storage[long_key][:long_count],
                    np.array(
                        (
                            generate_candle_from_one_minutes(
                                timeframe,
                                self.storage[short_key][short_count - dif:short_count],
                                True
                            ),
                        )
                    )
                ), axis=0
            )

    def get_current_candle(self, exchange, symbol, timeframe) -> np.ndarray:
        # no need to worry for forming candles when timeframe == 1m
        if timeframe == '1m':
            arr: DynamicNumpyArray = self.get_storage(exchange, symbol, '1m')
            if len(arr) == 0:
                return np.zeros((0, 6))
            else:
                return arr[-1]

        # other timeframes
        dif, long_key, short_key = self.forming_estimation(exchange, symbol, timeframe)
        long_count = len(self.get_storage(exchange, symbol, timeframe))
        short_count = len(self.get_storage(exchange, symbol, '1m'))

        # complete candle
        if dif == 0:
            if long_count == 0:
                return np.zeros((0, 6))
            else:
                return self.storage[long_key][-1]
        # generate forming
        else:
            return generate_candle_from_one_minutes(
                timeframe, self.storage[short_key][short_count - dif:short_count],
                True
            )



import pytest

import jesse.services.selectors as selectors
from jesse.config import config, reset_config
from jesse.enums import exchanges, timeframes, order_types, order_flags, order_roles
from jesse.exceptions import InvalidStrategy
from jesse.models import Position, Exchange
from jesse.routes import router
from jesse.services.broker import Broker
from jesse.store import store

position: Position = None
exchange: Exchange = None
broker: Broker = None


def set_up():
    """

    """
    reset_config()
    config['env']['exchanges'][exchanges.SANDBOX]['fee'] = 0
    config['env']['exchanges'][exchanges.SANDBOX]['starting_balance'] = 1000
    config['app']['trading_mode'] = 'backtest'
    config['app']['considering_exchanges'] = ['Sandbox']
    router.set_routes([(exchanges.SANDBOX, 'BTCUSD', '5m', 'Test19')])
    store.reset(True)

    global position
    global exchange
    global broker
    position = selectors.get_position(exchanges.SANDBOX, 'BTCUSD')
    position.current_price = 50
    exchange = selectors.get_exchange(exchanges.SANDBOX)
    broker = Broker(position, exchanges.SANDBOX, 'BTCUSD',
                    timeframes.MINUTE_5)


def set_up_with_fee():
    """

    """
    reset_config()
    config['env']['exchanges'][exchanges.SANDBOX]['fee'] = 0.002
    config['env']['exchanges'][exchanges.SANDBOX]['starting_balance'] = 1000
    config['app']['trading_mode'] = 'backtest'
    config['app']['considering_exchanges'] = ['Sandbox']
    router.set_routes([(exchanges.SANDBOX, 'BTCUSD', '5m', 'Test19')])
    store.reset(True)

    global position
    global exchange
    global broker
    position = selectors.get_position(exchanges.SANDBOX, 'BTCUSD')
    position.current_price = 50
    exchange = selectors.get_exchange(exchanges.SANDBOX)
    broker = Broker(position, exchanges.SANDBOX, 'BTCUSD',
                    timeframes.MINUTE_5)


def test_cancel_all_orders():
    set_up()

    # create 3 ACTIVE orders
    o1 = broker.buy_at(1, 40)
    o2 = broker.buy_at(1, 41)
    o3 = broker.buy_at(1, 42)
    assert o1.is_active
    assert o2.is_active
    assert o3.is_active

    # create 2 EXECUTED orders
    o4 = broker.buy_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    o5 = broker.buy_at_market(2)
    # fake it
    store.orders.execute_pending_market_orders()
    assert o4.is_executed
    assert o5.is_executed

    broker.cancel_all_orders()

    # ACTIVE orders should have been canceled
    assert o1.is_active is False
    assert o2.is_active is False
    assert o3.is_active is False
    assert o1.is_canceled is True
    assert o2.is_canceled is True
    assert o3.is_canceled is True
    # already-executed orders should have remain the same
    assert o4.is_executed
    assert o5.is_executed
    assert o4.is_canceled is False
    assert o5.is_canceled is False
def test_fee_calculation_when_opening_and_closing_positions():
    set_up_with_fee()

    # fee must be paid for opening the position
    assert position.qty == 0
    assert exchange.balance == 1000
    broker.sell_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.qty == -1
    assert exchange.balance == 949.9

    # fee must be cut again before increasing the balance when
    # closing the position
    broker.buy_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.is_close is True
    assert exchange.balance == 999.8


def test_fee_calculation_when_submitting_and_canceling_orders():
    set_up_with_fee()

    # fee must be cut from balance when submitting the order
    assert position.qty == 0
    assert exchange.balance == 1000
    order = broker.buy_at(2.5, 40)
    # fake it
    store.orders.execute_pending_market_orders()
    assert order.qty == 2.5
    assert order.is_active is True
    assert exchange.balance == 899.8

    # balance must be returned to the same(+fee)
    # when canceling open_position order
    order.cancel()
    assert exchange.balance == 1000

    # open a position
    broker.buy_at_market(2)
    # fake it
    store.orders.execute_pending_market_orders()
    assert exchange.balance == 899.8
    assert position.qty == 2
    # balance should be decreased when submitting increase_position order
    order = broker.buy_at(2.5, 40)
    # fake it
    store.orders.execute_pending_market_orders()
    assert round(exchange.balance, 1) == 799.6
    # balance must be returned to the same(+fee)
    # when canceling increase_position order
    order.cancel()
    assert exchange.balance == 899.8
    assert position.qty == 2

    # submitting a reduce_position should keep the balance the same
    order = broker.sell_at(2, 60)
    assert exchange.balance == 899.8
    # canceling the reduce_position order must not increase to balance
    order.cancel()
    assert exchange.balance == 899.8

    # however the balance must be incrased if the reduce_position
    # order gets executed
    assert position.qty == 2
    order = broker.sell_at(2, 60)
    assert exchange.balance == 899.8
    order.execute()
    assert position.is_close is True
    assert exchange.balance == 1019.56


def test_limit_orders():
    set_up()

    assert exchange.balance == 1000
    order = broker.buy_at(1, 40)
    assert order.price == 40
    assert order.qty == 1
    assert order.type == 'LIMIT'
    assert exchange.balance == 960
    broker.cancel_order(order.id)
    assert exchange.balance == 1000

    order = broker.sell_at(1, 60)
    assert order.price == 60
    assert order.qty == -1
    assert order.type == 'LIMIT'
    assert exchange.balance == 940
    broker.cancel_order(order.id)
    assert exchange.balance == 1000

    # validation: qty cannot be 0
    with pytest.raises(InvalidStrategy):
        broker.sell_at(0, 100)
        broker.buy_at(0, 100)


def test_market_orders_and_its_effects_on_position_and_exchange():
    set_up()

    # validation: qty cannot be 0
    with pytest.raises(InvalidStrategy):
        broker.sell_at_market(0)

    # validation: qty cannot be 0
    with pytest.raises(InvalidStrategy):
        broker.buy_at_market(0)

    assert position.qty == 0
    assert exchange.balance == 1000

    order = broker.sell_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.qty == -1
    assert exchange.balance == 950
    assert order.type == order_types.MARKET

    # increase position size
    broker.sell_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.qty == -2
    assert exchange.balance == 900

    # reduce position size
    broker.buy_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.qty == -1
    assert exchange.balance == 950

    # close position
    broker.buy_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.qty == 0
    assert exchange.balance == 1000

    # open long position
    broker.buy_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.qty == 1
    assert exchange.balance == 950

    # turn it into a short position
    broker.sell_at_market(3)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.qty == -2
    assert exchange.balance == 900


def test_opening_and_closing_position_with_stop():
    set_up()

    assert position.current_price == 50
    assert position.is_close is True
    assert exchange.balance == 1000
    # open position
    open_position_order = broker.start_profit_at('buy', 1, 60, order_roles.OPEN_POSITION)
    open_position_order.execute()
    assert position.is_open is True
    assert position.entry_price == 60
    assert position.qty == 1
    assert exchange.balance == 940

    # submit stop-loss order
    stop_loss_order = broker.stop_loss_at(1, 40, order_roles.CLOSE_POSITION)
    assert stop_loss_order.flag == order_flags.REDUCE_ONLY
    # balance should NOT have changed
    assert exchange.balance == 940
    # submit take-profit order also
    take_profit_order = broker.reduce_position_at(1, 80, order_roles.CLOSE_POSITION)
    assert take_profit_order.flag == order_flags.REDUCE_ONLY
    assert exchange.balance == 940

    # execute stop order
    stop_loss_order.execute()
    take_profit_order.cancel()
    assert position.is_close is True
    assert position.entry_price is None
    assert position.exit_price == 40
    assert exchange.balance == 980


def test_reduce_position_order():
    set_up()

    assert position.current_price == 50
    assert exchange.balance == 1000
    broker.buy_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.qty == 1
    assert exchange.balance == 950

    # $50 was spent at opening position
    order = broker.reduce_position_at(1, 60)
    assert order.price == 60
    assert order.qty == -1
    assert order.type == 'LIMIT'
    assert order.flag == order_flags.REDUCE_ONLY

    # neither opening nor canceling reduce position will affect balance
    assert exchange.balance == 950
    broker.cancel_order(order.id)
    assert exchange.balance == 950

    # but it's execution will
    order = broker.reduce_position_at(1, 60)
    order.execute()
    assert exchange.balance == 1010
    assert position.is_close is True

    # validation: qty cannot be 0
    with pytest.raises(InvalidStrategy):
        broker.reduce_position_at(0, 100)


def test_start_profit():
    set_up()

    assert position.current_price == 50
    assert position.is_close is True
    assert exchange.balance == 1000

    order = broker.start_profit_at('buy', 1, 60)
    assert exchange.balance == 940
    assert order.type == order_types.STOP
    assert order.qty == 1

    order.cancel()
    assert exchange.balance == 1000

    order = broker.start_profit_at('buy', 1, 60)
    assert exchange.balance == 940
    order.execute()
    assert exchange.balance == 940
    assert position.qty == 1
    assert position.entry_price == 60

    # validation: qty cannot be 0
    with pytest.raises(InvalidStrategy):
        broker.start_profit_at('buy', 0, 100)


def test_stop_loss():
    set_up()

    assert position.current_price == 50
    assert position.is_close is True
    assert exchange.balance == 1000
    # open position
    broker.buy_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert position.is_open is True
    assert position.entry_price == 50
    assert position.qty == 1
    assert exchange.balance == 950

    # submit stop-loss order
    order = broker.stop_loss_at(1, 40)
    assert order.type == order_types.STOP
    assert order.price == 40
    assert order.qty == -1
    assert order.side == 'sell'
    assert order.flag == order_flags.REDUCE_ONLY
    # balance should NOT have changed
    assert exchange.balance == 950

    # execute stop order
    order.execute()
    assert position.is_close == True
    assert position.entry_price == None
    assert position.exit_price == 40
    assert exchange.balance == 990

    # make sure stop_loss order cannot open
    # position even in the other direction
    # because of its REDUCE_ONLY flag
    broker.buy_at_market(1)
    # fake it
    store.orders.execute_pending_market_orders()
    assert exchange.balance == 940
    assert position.qty == 1
    order = broker.stop_loss_at(2, 40)
    order.execute()
    assert position.qty == 0
    assert exchange.balance == 940

    # validation: qty cannot be 0
    with pytest.raises(InvalidStrategy):
        broker.stop_loss_at(0, 100)





import jesse.helpers as jh
from jesse.config import config, reset_config
from jesse.models import CompletedTrade
from jesse.store import store


def no_fee():
    """

    """
    print(config)
    config['env']['exchanges']['Sandbox']['fee'] = 0


def set_up():
    """

    """
    reset_config()
    store.reset(True)


def test_can_add_trade_to_store():
    set_up()

    assert store.completed_trades.trades == []

    trade = CompletedTrade({
        'type': 'long',
        'exchange': 'Sandbox',
        'entry_price': 10,
        'exit_price': 20,
        'take_profit_at': 20,
        'stop_loss_at': 5,
        'qty': 1,
        'orders': [],
        'symbol': 'BTCUSD',
        'opened_at': 1552309186171,
        'closed_at': 1552309186171 + 60000
    })

    store.completed_trades.add_trade(trade)
    assert store.completed_trades.trades == [trade]
    store.reset()
    assert store.completed_trades.trades == []
def test_holding_period():
    trade = CompletedTrade({
        'type': 'long',
        'exchange': 'Sandbox',
        'entry_price': 10,
        'exit_price': 20,
        'take_profit_at': 20,
        'stop_loss_at': 5,
        'qty': 1,
        'orders': [],
        'symbol': 'BTCUSD',
        'opened_at': 1552309186171,
        'closed_at': 1552309186171 + 60000
    })

    # 1 minute == 60 seconds
    assert trade.holding_period == 60


def test_PNL_percentage():
    no_fee()

    trade = CompletedTrade({
        'type': 'long',
        'exchange': 'Sandbox',
        'entry_price': 10,
        'exit_price': 12,
        'take_profit_at': 20,
        'stop_loss_at': 5,
        'qty': 1,
        'orders': [],
        'symbol': 'BTCUSD',
        'opened_at': jh.now(),
        'closed_at': jh.now()
    })
    assert trade.PNL_percentage == 20


def test_PNL_with_fee():
    # set fee (0.20%)
    config['env']['exchanges']['Sandbox']['fee'] = 0.002

    trade = CompletedTrade({
        'type': 'long',
        'exchange': 'Sandbox',
        'entry_price': 10,
        'exit_price': 20,
        'take_profit_at': 20,
        'stop_loss_at': 5,
        'qty': 1,
        'orders': [],
        'symbol': 'BTCUSD',
        'opened_at': jh.now(),
        'closed_at': jh.now()
    })

    assert trade.fee == 0.06
    assert trade.PNL == 9.94


def test_PNL_without_fee():
    no_fee()

    trade = CompletedTrade({
        'type': 'long',
        'exchange': 'Sandbox',
        'entry_price': 10,
        'exit_price': 20,
        'take_profit_at': 20,
        'stop_loss_at': 5,
        'qty': 1,
        'orders': [],
        'symbol': 'BTCUSD',
        'opened_at': jh.now(),
        'closed_at': jh.now()
    })
    assert trade.PNL == 10


def test_R():
    no_fee()

    trade = CompletedTrade({
        'type': 'long',
        'exchange': 'Sandbox',
        'entry_price': 10,
        'exit_price': 12,
        'take_profit_at': 20,
        'stop_loss_at': 5,
        'qty': 1,
        'orders': [],
        'symbol': 'BTCUSD',
        'opened_at': jh.now(),
        'closed_at': jh.now()
    })
    assert trade.risk_reward_ratio == 2


def test_risk_percentage():
    no_fee()

    trade = CompletedTrade({
        'type': 'long',
        'exchange': 'Sandbox',
        'entry_price': 10,
        'exit_price': 12,
        'take_profit_at': 20,
        'stop_loss_at': 5,
        'qty': 1,
        'orders': [],
        'symbol': 'BTCUSD',
        'opened_at': jh.now(),
        'closed_at': jh.now()
    })
    assert trade.risk_percentage == round((((10 - 5) / 1) * 10), 2)


def test_trade_size():
    trade = CompletedTrade({
        'type': 'long',
        'exchange': 'Sandbox',
        'entry_price': 10,
        'exit_price': 20,
        'take_profit_at': 20,
        'stop_loss_at': 5,
        'qty': 1,
        'orders': [],
        'symbol': 'BTCUSD',
        'opened_at': jh.now(),
        'closed_at': jh.now()
    })

    assert trade.size == 10





from collections import namedtuple

import numpy as np
import talib

from jesse.helpers import get_candle_source

MACDEXT = namedtuple('MACDEXT', ['macd', 'signal', 'hist'])


def macdext(candles: np.ndarray, fastperiod=12, fastmatype=0, slowperiod=26, slowmatype=0, signalperiod=9,
            signalmatype=0, source_type="close", sequential=False) -> MACDEXT:
    """
    MACDEXT - MACD with controllable MA type

    :param candles: np.ndarray
    :param fastperiod: int - default: 12
    :param fastmatype: int - default: 0
    :param slow_period: int - default: 26
    :param slowmatype: int - default: 0
    :param signal_period: int - default: 9
    :param signalmatype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: MACDEXT(macd, signal, hist)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    macd, macdsignal, macdhist = talib.MACDEXT(source, fastperiod=fastperiod, fastmatype=fastmatype,
                                               slowperiod=slowperiod, slowmatype=slowmatype,
                                               signalperiod=signalperiod, signalmatype=signalmatype)

    if sequential:
        return MACDEXT(macd, macdsignal, macdhist)
    else:
        return MACDEXT(macd[-1], macdsignal[-1], macdhist[-1])



from typing import Union

import numpy as np
import tulipy as ti

from jesse.helpers import get_candle_source


def vidya(candles: np.ndarray, short_period=2,  long_period=5,  alpha=0.2, source_type="close", sequential=False) -> Union[float, np.ndarray]:
    """
    VIDYA - Variable Index Dynamic Average

    :param candles: np.ndarray
    :param short_period: int - default: 2
    :param long_period: int - default: 5
    :param alpha: float - default: 0.2
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    res = ti.vidya(np.ascontiguousarray(source), short_period=short_period, long_period=long_period, alpha=alpha)

    return np.concatenate((np.full((candles.shape[0] - res.shape[0]), np.nan), res), axis=0) if sequential else res[-1]



from typing import Union

import numpy as np
import talib

from jesse.helpers import get_candle_source


def midpoint(candles: np.ndarray, period=14, source_type="close", sequential=False) -> Union[float, np.ndarray]:
    """
    MIDPOINT - MidPoint over period

    :param candles: np.ndarray
    :param period: int - default=14
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    res = talib.MIDPOINT(source, timeperiod=period)

    if sequential:
        return res
    else:
        return None if np.isnan(res[-1]) else res[-1]



import math
from typing import Union

import numpy as np

from jesse.helpers import get_candle_source


def supersmoother(candles: np.ndarray, cutoff=14, source_type="close", sequential=False) -> Union[float, np.ndarray]:
    """
    Super Smoother Filter 2pole Butterworth
    This indicator was described by John F. Ehlers

    :param candles: np.ndarray
    :param cutoff: int - default=14
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """

    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    source = source.flatten()
    N = len(source)
    source = source[~np.isnan(source)]
    to_fill = N - len(source)

    PI = math.pi
    f = (math.sqrt(2) * PI) / cutoff
    a = math.exp(-f)
    c2 = 2 * a * math.cos(f)
    c3 = -a * a
    c1 = 1 - c2 - c3

    src = np.insert(source, 0, 0)
    coeff = np.array([0.5 * c1, 0.5 * c1, c2, c3])
    fil = np.zeros(2 + len(source))
    for i in range(len(source)):
        val = np.array([src[i,], src[i + 1,], fil[i + 1], fil[i]])
        fil[2 + i] = np.dot(coeff, val)
    if to_fill != 0:
        out = np.insert(fil[2:], 0, np.repeat(np.nan, to_fill))
    else:
        out = fil[2:]

    res = out

    if sequential:
        return res
    else:
        return None if np.isnan(res[-1]) else res[-1]



import arrow
import click
import numpy as np

import jesse.helpers as jh


def generate_candle_from_one_minutes(timeframe,
                                     candles: np.ndarray,
                                     accept_forming_candles=False):
    """

    :param timeframe:
    :param candles:
    :param accept_forming_candles:
    :return:
    """
    if len(candles) == 0:
        raise ValueError('No candles were passed')

    if not accept_forming_candles and len(candles) != jh.timeframe_to_one_minutes(timeframe):
        raise ValueError(
            'Sent only {} candles but {} is required to create a "{}" candle.'.format(
                len(candles), jh.timeframe_to_one_minutes(timeframe),
                timeframe
            )
        )

    return np.array([
        candles[0][0],
        candles[0][1],
        candles[-1][2],
        candles[:, 3].max(),
        candles[:, 4].min(),
        candles[:, 5].sum(),
    ])


def print_candle(candle, is_partial, symbol):
    """

    :param candle:
    :param is_partial:
    :param symbol:
    :return:
    """
    if jh.should_execute_silently():
        return

    if is_bullish(candle) and is_partial is True:
        candle_form = click.style('  ==', fg='green')
    elif is_bullish(candle) and is_partial is False:
        candle_form = click.style('====', bg='green')
    elif is_bearish(candle) and is_partial is True:
        candle_form = click.style('  ==', fg='red')
    else:
        candle_form = click.style('====', bg='red')

    if is_bullish(candle):
        candle_info = click.style(' {} | {} | {} | {} | {} | {} | {}'.format(
            symbol, str(arrow.get(candle[0] / 1000))[:-9], candle[1], candle[2],
            candle[3], candle[4], round(candle[5], 2)),
            fg='green')
    else:
        candle_info = click.style(' {} | {} | {} | {} | {} | {} | {}'.format(
            symbol, str(arrow.get(candle[0] / 1000))[:-9], candle[1], candle[2],
            candle[3], candle[4], round(candle[5], 2)),
            fg='red')

    print(candle_form + candle_info)


def is_bullish(candle: np.ndarray) -> bool:
    """

    :param candle:
    :return:
    """
    return candle[2] >= candle[1]


def is_bearish(candle: np.ndarray) -> bool:
    """

    :param candle:
    :return:
    """
    return candle[2] < candle[1]


def candle_includes_price(candle, price) -> bool:
    """

    :param candle:
    :param price:
    :return:
    """
    return (price >= candle[4]) and (price <= candle[3])


def split_candle(candle: np.ndarray, price: float) -> tuple:
    """
    splits a single candle into two candles: earlier + later

    :param candle: np.ndarray
    :param price: float

    :return: tuple
    """
    timestamp = candle[0]
    o = candle[1]
    c = candle[2]
    h = candle[3]
    l = candle[4]
    v = candle[5]

    if is_bullish(candle) and l < price < o:
        return np.array([
            timestamp, o, price, o, price, v
        ]), np.array([
            timestamp, price, c, h, l, v
        ])
    elif price == o:
        return candle, candle
    elif is_bearish(candle) and o < price < h:
        return np.array([
            timestamp, o, price, price, o, v
        ]), np.array([
            timestamp, price, c, h, l, v
        ])
    elif is_bearish(candle) and l < price < c:
        return np.array([
            timestamp, o, price, h, price, v
        ]), np.array([
            timestamp, price, c, c, l, v
        ])
    elif is_bullish(candle) and c < price < h:
        return np.array([
            timestamp, o, price, price, l, v
        ]), np.array([
            timestamp, price, c, h, c, v
        ]),
    elif is_bearish(candle) and price == c:
        return np.array([
            timestamp, o, c, h, c, v
        ]), np.array([
            timestamp, price, price, price, l, v
        ])
    elif is_bullish(candle) and price == c:
        return np.array([
            timestamp, o, c, c, l, v
        ]), np.array([
            timestamp, price, price, h, price, v
        ])
    elif is_bearish(candle) and price == h:
        return np.array([
            timestamp, o, h, h, o, v
        ]), np.array([
            timestamp, h, c, h, l, v
        ])
    elif is_bullish(candle) and price == l:
        return np.array([
            timestamp, o, l, o, l, v
        ]), np.array([
            timestamp, l, c, h, l, v
        ])
    elif is_bearish(candle) and price == l:
        return np.array([
            timestamp, o, l, h, l, v
        ]), np.array([
            timestamp, l, c, c, l, v
        ])
    elif is_bullish(candle) and price == h:
        return np.array([
            timestamp, o, h, h, l, v
        ]), np.array([
            timestamp, h, c, h, c, v
        ])
    elif is_bearish(candle) and c < price < o:
        return np.array([
            timestamp, o, price, h, price, v
        ]), np.array([
            timestamp, price, c, price, l, v
        ])
    elif is_bullish(candle) and o < price < c:
        return np.array([
            timestamp, o, price, price, l, v
        ]), np.array([
            timestamp, price, c, h, price, v
        ])



from typing import Union

import numpy as np
import talib

from jesse.helpers import get_candle_source


def apo(candles: np.ndarray, fastperiod=12, slowperiod=26, matype=0, source_type="close", sequential=False) -> Union[
    float, np.ndarray]:
    """
    APO - Absolute Price Oscillator

    :param candles: np.ndarray
    :param fastperiod: int - default: 12
    :param slowperiod: int - default: 26
    :param matype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)

    res = talib.APO(source, fastperiod=fastperiod, slowperiod=slowperiod, matype=matype)

    return res if sequential else res[-1]



from jesse.strategies import Strategy


# test_is_reduced
class Test43(Strategy):
    def should_long(self) -> bool:
        return self.index == 0

    def should_short(self) -> bool:
        return False

    def go_long(self):
        self.buy = 1, 2
        self.take_profit = [
            (.5, 6),
            (.5, 10),
        ]

    def update_position(self):
        if self.position.qty == 1:
            assert not self.is_increased
            assert not self.is_reduced
        elif self.position.qty == .5:
            assert not self.is_increased
            assert self.is_reduced

    def go_short(self):
        pass

    def should_cancel(self):
        return False



from collections import namedtuple

import numpy as np
import talib

from jesse.helpers import get_candle_source

MACD = namedtuple('MACD', ['macd', 'signal', 'hist'])


def macd(candles: np.ndarray, fastperiod=12, slowperiod=26, signalperiod=9, source_type="close",
         sequential=False) -> MACD:
    """
    MACD - Moving Average Convergence/Divergence

    :param candles: np.ndarray
    :param fastperiod: int - default: 12
    :param slow_period: int - default: 26
    :param signal_period: int - default: 9
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: MACD(macd, signal, hist)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    macd, macdsignal, macdhist = talib.MACD(source, fastperiod=fastperiod, slowperiod=slowperiod,
                                            signalperiod=signalperiod)

    if sequential:
        return MACD(macd, macdsignal, macdhist)
    else:
        return MACD(macd[-1], macdsignal[-1], macdhist[-1])



from typing import Union

import numpy as np
import talib

from jesse.helpers import get_candle_source


def cc(candles: np.ndarray, wma_period=10, roc_short_period=11, roc_long_period=14, source_type="close", sequential=False) -> Union[float, np.ndarray]:
    """
    CC - Coppock Curve

    :param candles: np.ndarray
    :param wma_period: int - default: 10
    :param roc_short_period: int - default: 11
    :param roc_long_period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    res = talib.WMA(talib.ROC(source, timeperiod=roc_long_period) + talib.ROC(source, timeperiod=roc_short_period), timeperiod=wma_period)

    return res if sequential else res[-1]



# silent (pandas) warnings
import warnings

import numpy as np
import pandas as pd

import jesse.helpers as jh
from jesse.config import config
from jesse.routes import router
from jesse.services import statistics as stats
from jesse.services.candle import is_bullish
from jesse.store import store

warnings.filterwarnings("ignore")


def positions():
    array = []

    # headers
    array.append(['type', 'strategy', 'symbol', 'opened at', 'qty', 'entry', 'current price', 'PNL (%)'])

    for p in store.positions.storage:
        pos = store.positions.storage[p]

        if pos.pnl_percentage > 0:
            pnl_color = 'green'
        elif pos.pnl_percentage < 0:
            pnl_color = 'red'
        else:
            pnl_color = 'black'

        if pos.type == 'long':
            type_color = 'green'
        elif pos.type == 'short':
            type_color = 'red'
        else:
            type_color = 'black'

        array.append(
            [
                jh.color(pos.type, type_color),
                pos.strategy.name,
                pos.symbol,
                '' if pos.is_close else '{} ago'.format(jh.readable_duration((jh.now() - pos.opened_at) / 1000, 3)),
                pos.qty if abs(pos.qty) > 0 else None,
                pos.entry_price,
                pos.current_price,
                '' if pos.is_close else '{} ({}%)'.format(
                    jh.color(str(round(pos.pnl, 2)), pnl_color),
                    jh.color(str(round(pos.pnl_percentage, 4)), pnl_color)
                ),
            ]
        )

    return array


def candles():
    array = []
    candle_keys = []

    # add routes
    for e in router.routes:
        if e.strategy is None:
            return

        candle_keys.append({
            'exchange': e.exchange,
            'symbol': e.symbol,
            'timeframe': e.timeframe
        })

    # add extra_routes
    for e in router.extra_candles:
        candle_keys.append({
            'exchange': e[0],
            'symbol': e[1],
            'timeframe': e[2]
        })

    # headers
    array.append(['exchange-symbol-timeframe', 'timestamp', 'open', 'close', 'high', 'low'])

    for k in candle_keys:
        try:
            current_candle = store.candles.get_current_candle(k['exchange'], k['symbol'], k['timeframe'])
            green = is_bullish(current_candle)
            bold = k['symbol'] in config['app']['trading_symbols'] and k['timeframe'] in config['app'][
                'trading_timeframes']
            key = jh.key(k['exchange'], k['symbol'], k['timeframe'])
            array.append(
                [
                    jh.style(key, 'underline' if bold else None),
                    jh.color(jh.timestamp_to_time(current_candle[0]), 'green' if green else 'red'),
                    jh.color(str(current_candle[1]), 'green' if green else 'red'),
                    jh.color(str(current_candle[2]), 'green' if green else 'red'),
                    jh.color(str(current_candle[3]), 'green' if green else 'red'),
                    jh.color(str(current_candle[4]), 'green' if green else 'red'),
                ]
            )
        except IndexError:
            return
        except Exception:
            raise

    return array


def livetrade():
    # sum up balance of all trading exchanges
    starting_balance = 0
    current_balance = 0
    for e in store.exchanges.storage:
        starting_balance += store.exchanges.storage[e].starting_balance
        current_balance += store.exchanges.storage[e].balance
    starting_balance = round(starting_balance, 2)
    current_balance = round(current_balance, 2)

    arr = [
        ['started/current balance', '{}/{}'.format(starting_balance, current_balance)],
        ['started at', jh.get_arrow(store.app.starting_time).humanize()],
        ['current time', jh.timestamp_to_time(jh.now())[:19]],
        ['errors/info', '{}/{}'.format(len(store.logs.errors), len(store.logs.info))],
        ['active orders', store.orders.count_all_active_orders()],
        ['open positions', store.positions.count_open_positions()]
    ]

    # short trades summary
    if len(store.completed_trades.trades):
        df = pd.DataFrame.from_records([t.to_dict() for t in store.completed_trades.trades])
        total = len(df)
        winning_trades = df.loc[df['PNL'] > 0]
        losing_trades = df.loc[df['PNL'] < 0]
        pnl = round(df['PNL'].sum(), 2)
        pnl_percentage = round((pnl / starting_balance) * 100, 2)

        arr.append(['total/winning/losing trades', '{}/{}/{}'.format(total, len(winning_trades), len(losing_trades))])
        arr.append(['PNL (%)', '${} ({}%)'.format(pnl, pnl_percentage)])

    if config['app']['debug_mode']:
        arr.append(['debug mode', config['app']['debug_mode']])

    if config['app']['is_test_driving']:
        arr.append(['Test Drive', config['app']['is_test_driving']])

    return arr


def portfolio_metrics():
    data = stats.trades(store.completed_trades.trades, store.app.daily_balance)

    metrics = [
        ['Total Closed Trades', data['total']],
        ['Total Net Profit',
         '{} ({})'.format(round(data['net_profit'], 4), str(data['net_profit_percentage']) + '%')],
        ['Starting => Finishing Balance',
         '{} => {}'.format(round(data['starting_balance'], 2), round(data['finishing_balance'], 2))],
        ['Total Open Trades', data['total_open_trades']],
        ['Open PL', round(data['open_pl'], 2)],
        ['Total Paid Fees', round(data['fee'], 2)],
        ['Max Drawdown', '{}%'.format(data['max_drawdown'])],
        ['Annual Return', '{}%'.format(data['annual_return'])],
        ['Expectancy',
         '{} ({})'.format(round(data['expectancy'], 2), str(round(data['expectancy_percentage'], 2)) + '%')],
        ['Avg Win | Avg Loss', '{} | {}'.format(round(data['average_win'], 2), round(data['average_loss'], 2))],
        ['Ratio Avg Win / Avg Loss', round(data['ratio_avg_win_loss'], 2)],
        ['Percent Profitable', str(round(data['win_rate'] * 100)) + '%'],
        ['Longs | Shorts', '{}% | {}%'.format(round(data['longs_percentage']), round(data['short_percentage']))],
        ['Avg Holding Time', jh.readable_duration(data['average_holding_period'], 3)],
        ['Winning Trades Avg Holding Time',
         np.nan if np.isnan(data['average_winning_holding_period']) else jh.readable_duration(
             data['average_winning_holding_period'], 3)],
        ['Losing Trades Avg Holding Time',
         np.nan if np.isnan(data['average_losing_holding_period']) else jh.readable_duration(
             data['average_losing_holding_period'], 3)]
    ]

    if jh.get_config('env.metrics.sharpe_ratio', True):
        metrics.append(['Sharpe Ratio', data['sharpe_ratio']])
    if jh.get_config('env.metrics.calmar_ratio', False):
        metrics.append(['Calmar Ratio', data['calmar_ratio']])
    if jh.get_config('env.metrics.sortino_ratio', False):
        metrics.append(['Sortino Ratio', data['sortino_ratio']])
    if jh.get_config('env.metrics.omega_ratio', False):
        metrics.append(['Omega Ratio', data['omega_ratio']])
    if jh.get_config('env.metrics.winning_streak', False):
        metrics.append(['Winning Streak', data['winning_streak']])
    if jh.get_config('env.metrics.losing_streak', False):
        metrics.append(['Losing Streak', data['losing_streak']])
    if jh.get_config('env.metrics.largest_winning_trade', False):
        metrics.append(['Largest Winning Trade', data['largest_winning_trade']])
    if jh.get_config('env.metrics.largest_losing_trade', False):
        metrics.append(['Largest Losing Trade', data['largest_losing_trade']])
    if jh.get_config('env.metrics.total_winning_trades', False):
        metrics.append(['Total Winning Trades', data['total_winning_trades']])
    if jh.get_config('env.metrics.total_losing_trades', False):
        metrics.append(['Total Losing Trades', data['total_losing_trades']])

    return metrics


def info():
    array = []

    for w in store.logs.info[::-1][0:5]:
        array.append(
            [jh.timestamp_to_time(w['time'])[11:19],
             (w['message'][:70] + '..') if len(w['message']) > 70 else w['message']])

    return array


def watch_list():
    # only support one route
    if len(router.routes) > 1:
        return None

    strategy = router.routes[0].strategy

    # don't if the strategy hasn't been initiated yet
    if not store.candles.is_initiated:
        return None

    watch_list_array = strategy.watch_list()

    return watch_list_array if len(watch_list_array) else None


def errors():
    array = []

    for w in store.logs.errors[::-1][0:5]:
        array.append([jh.timestamp_to_time(w['time'])[11:19],
                      (w['message'][:70] + '..') if len(w['message']) > 70 else w['message']])

    return array


def orders():
    array = []

    # headers
    array.append(['symbol', 'side', 'type', 'qty', 'price', 'flag', 'status', 'created_at'])

    route_orders = []
    for r in router.routes:
        r_orders = store.orders.get_orders(r.exchange, r.symbol)
        for o in r_orders:
            route_orders.append(o)

    if not len(route_orders):
        return None

    route_orders.sort(key=lambda x: x.created_at, reverse=False)

    for o in route_orders[::-1][0:5]:
        array.append([
            o.symbol if o.is_active else jh.color(o.symbol, 'gray'),
            jh.color(o.side, 'red') if o.side == 'sell' else jh.color(o.side, 'green'),
            o.type if o.is_active else jh.color(o.type, 'gray'),
            o.qty if o.is_active else jh.color(str(o.qty), 'gray'),
            o.price if o.is_active else jh.color(str(o.price), 'gray'),
            o.flag if o.is_active else jh.color(o.flag, 'gray'),
            o.status if o.is_active else jh.color(o.status, 'gray'),
            jh.timestamp_to_time(o.created_at)[:19] if o.is_active else jh.color(
                jh.timestamp_to_time(o.created_at)[:19], 'gray'),
        ])

    return array



import math
from typing import Union

import numpy as np

from jesse.helpers import get_candle_source


def gauss(candles: np.ndarray, period=14, poles=4, source_type="close", sequential=False) -> Union[float, np.ndarray]:
    """
    Gaussian Filter

    :param candles: np.ndarray
    :param period: int - default=14
    :param poles: int - default=4
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """

    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    N = len(source)
    source = source[~np.isnan(source)]
    to_fill = N - len(source)
    PI = math.pi
    beta = (1 - math.cos(2 * PI / period)) / (math.pow(2, 1 / poles) - 1)
    alpha = -beta + math.sqrt(math.pow(beta, 2) + 2 * beta)

    fil = np.zeros(poles + len(source))
    if poles == 1:
        coeff = np.array([alpha, (1 - alpha)])
    elif poles == 2:
        coeff = np.array([alpha ** 2, 2 * (1 - alpha), -(1 - alpha) ** 2])
    elif poles == 3:
        coeff = np.array([alpha ** 3, 3 * (1 - alpha), -3 * (1 - alpha) ** 2, (1 - alpha) ** 3])
    elif poles == 4:
        coeff = np.array([alpha ** 4, 4 * (1 - alpha), -6 * (1 - alpha) ** 2, 4 * (1 - alpha) ** 3, -(1 - alpha) ** 4])

    for i in range(len(source)):
        if poles == 1:
            val = np.array([source[i].item(), fil[i]])
        elif poles == 2:
            val = np.array([source[i].item(), fil[1 + i], fil[i]])
        elif poles == 3:
            val = np.array([source[i].item(), fil[2 + i], fil[1 + i], fil[i]])
        elif poles == 4:
            val = np.array([source[i].item(), fil[3 + i], fil[2 + i], fil[1 + i], fil[i]])

        fil[poles + i] = np.dot(coeff, val)

    if to_fill != 0:
        res = np.insert(fil[poles:], 0, np.repeat(np.nan, to_fill))
    else:
        res = fil[poles:]

    if sequential:
        return res
    else:
        return None if np.isnan(res[-1]) else res[-1]



import jesse.helpers as jh
from jesse.enums import sides, order_flags
from jesse.exceptions import OrderNotAllowed, InvalidStrategy
from jesse.models import Order
from jesse.models import Position


class Broker:
    """

    """
    def __init__(self, position, exchange, symbol, timeframe):
        self.position: Position = position
        self.symbol = symbol
        self.timeframe = timeframe
        self.exchange = exchange
        from jesse.services.api import api
        self.api = api

    @staticmethod
    def _validate_qty(qty):
        if qty == 0:
            raise InvalidStrategy('qty cannot be 0')

    def sell_at_market(self, qty, role=None) -> Order:
        """

        :param qty:
        :param role:
        :return:
        """
        self._validate_qty(qty)

        return self.api.market_order(
            self.exchange,
            self.symbol,
            abs(qty),
            self.position.current_price,
            sides.SELL,
            role, []
        )

    def sell_at(self, qty, price, role=None) -> Order:
        """

        :param qty:
        :param price:
        :param role:
        :return:
        """
        self._validate_qty(qty)

        if price < 0:
            raise ValueError('price cannot be negative.')

        if price <= self.position.current_price:
            raise OrderNotAllowed(
                'Cannot LIMIT sell at ${} when current_price is ${}'.format(
                    price,
                    self.position.current_price
                )
            )

        return self.api.limit_order(
            self.exchange,
            self.symbol,
            abs(qty),
            price,
            sides.SELL,
            role,
            []
        )

    def buy_at_market(self, qty, role=None) -> Order:
        """

        :param qty:
        :param role:
        :return:
        """
        self._validate_qty(qty)

        return self.api.market_order(
            self.exchange,
            self.symbol,
            abs(qty),
            self.position.current_price,
            sides.BUY,
            role,
            []
        )

    def buy_at(self, qty, price, role=None) -> Order:
        """

        :param qty:
        :param price:
        :param role:
        :return:
        """
        self._validate_qty(qty)

        if price < 0:
            raise ValueError('price cannot be negative.')

        if price >= self.position.current_price:
            raise OrderNotAllowed(
                'Cannot LIMIT buy at ${} when current_price is ${}'.format(
                    price,
                    self.position.current_price
                )
            )

        return self.api.limit_order(
            self.exchange,
            self.symbol,
            abs(qty),
            price,
            sides.BUY,
            role,
            []
        )

    def reduce_position_at(self, qty, price, role=None) -> Order:
        """

        :param qty:
        :param price:
        :param role:
        :return:
        """
        self._validate_qty(qty)

        qty = abs(qty)

        # validation
        if price < 0:
            raise ValueError('price cannot be negative.')

        # validation
        if self.position.is_close:
            raise OrderNotAllowed(
                'Cannot submit a reduce_position order when there is not open position'
            )

        side = jh.opposite_side(jh.type_to_side(self.position.type))

        # validation
        if side == 'buy' and price >= self.position.current_price:
            raise OrderNotAllowed(
                'Cannot reduce (via LIMIT) buy at ${} when current_price is ${}'.format(
                    price,
                    self.position.current_price
                )
            )
        # validation
        if side == 'sell' and price <= self.position.current_price:
            raise OrderNotAllowed(
                'Cannot reduce (via LIMIT) sell at ${} when current_price is ${}'.format(
                    price,
                    self.position.current_price
                )
            )

        return self.api.limit_order(
            self.exchange,
            self.symbol,
            qty,
            price,
            side,
            role,
            [order_flags.REDUCE_ONLY]
        )

    def start_profit_at(self, side, qty, price, role=None) -> Order:
        """

        :param side:
        :param qty:
        :param price:
        :param role:
        :return:
        """
        self._validate_qty(qty)

        if price < 0:
            raise ValueError('price cannot be negative.')

        if side == 'buy' and price < self.position.current_price:
            raise OrderNotAllowed(
                'A buy start_profit({}) order must have a price higher than current_price({}).'.format(
                    price,
                    self.position.current_price
                )
            )
        if side == 'sell' and price > self.position.current_price:
            raise OrderNotAllowed(
                'A sell start_profit({}) order must have a price lower than current_price({}).'.format(
                    price,
                    self.position.current_price
                )
            )

        return self.api.stop_order(
            self.exchange,
            self.symbol,
            abs(qty),
            price,
            side,
            role,
            []
        )

    def stop_loss_at(self, qty, price, role=None) -> Order:
        """

        :param qty:
        :param price:
        :param role:
        :return:
        """
        self._validate_qty(qty)

        side = jh.opposite_side(jh.type_to_side(self.position.type))

        if price < 0:
            raise ValueError('price cannot be negative.')

        if side == 'buy' and price < self.position.current_price:
            raise OrderNotAllowed(
                'Cannot submit a buy stop at {} when current price is {}'.format(
                    price,
                    self.position.current_price
                )
            )
        if side == 'sell' and price > self.position.current_price:
            raise OrderNotAllowed(
                'Cannot submit a sell stop at {} when current price is {}.'.format(
                    price,
                    self.position.current_price
                )
            )

        return self.api.stop_order(
            self.exchange,
            self.symbol,
            abs(qty),
            price,
            side,
            role,
            [order_flags.REDUCE_ONLY]
        )

    def cancel_all_orders(self):
        return self.api.cancel_all_orders(self.exchange, self.symbol)

    def cancel_order(self, order_id: str):
        """

        :param order_id:
        :return:
        """
        return self.api.cancel_order(self.exchange, self.symbol, order_id)



import jesse.helpers as jh
import jesse.services.selectors as selectors
import jesse.utils as ju
from jesse.config import config
from jesse.enums import trade_types, order_types, order_flags
from jesse.exceptions import EmptyPosition, OpenPositionError
from jesse.models import Order
from jesse.services import logger, notifier


class Position:
    """

    """
    def __init__(self, exchange_name, symbol, attributes=None):
        self.id = jh.generate_unique_id()
        self.entry_price = None
        self.exit_price = None
        self.current_price = None
        self.qty = 0
        self.opened_at = None
        self.closed_at = None

        if attributes is None:
            attributes = {}

        self.exchange_name = exchange_name
        self.exchange = selectors.get_exchange(self.exchange_name)
        self.symbol = symbol
        self.strategy = None

        for a in attributes:
            setattr(self, a, attributes[a])

    @property
    def value(self):
        """
        The value of open position

        :return: float
        """
        return abs(self.current_price * self.qty)

    @property
    def type(self):
        """
        The type of open position - long, short, or close

        :return: str
        """
        if self.qty > 0:
            return 'long'
        if self.qty < 0:
            return 'short'

        return 'close'

    @property
    def pnl_percentage(self):
        """
        The PNL% of the position

        :return: float
        """
        if self.qty == 0:
            return 0

        exit_price = self.current_price if self.exit_price is None else self.exit_price

        return jh.estimate_PNL_percentage(self.qty, self.entry_price, exit_price, self.type)

    @property
    def pnl(self):
        """
        The PNL of the position

        :return: float
        """
        if self.qty == 0:
            return 0

        diff = self.value - abs(self.entry_price * self.qty)

        return -diff if self.type == 'short' else diff

    @property
    def is_open(self):
        """
        Is the current position open?

        :return: bool
        """
        return self.qty != 0

    @property
    def is_close(self):
        """
        Is the current position close?

        :return: bool
        """
        return self.qty == 0

    def _close(self, close_price):
        if self.is_open is False:
            raise EmptyPosition('The position is already closed.')

        # just to prevent confusion
        close_qty = abs(self.qty)

        estimated_profit = jh.estimate_PNL(
            close_qty, self.entry_price,
            close_price, self.type
        )
        entry = self.entry_price
        trade_type = self.type
        self.exit_price = close_price

        if self.exchange:
            self.exchange.increase_balance(
                self, close_qty * self.entry_price + estimated_profit
            )
        self.qty = 0
        self.entry_price = None
        self.closed_at = jh.now()

        info_text = 'CLOSED {} position: {}, {}. PNL: ${}, entry: {}, exit: {}'.format(
            trade_type, self.exchange_name, self.symbol, round(estimated_profit, 2), entry, close_price
        )

        if jh.is_debuggable('position_closed'):
            logger.info(info_text)

        if jh.is_live() and config['env']['notifications']['events']['updated_position']:
            notifier.notify(info_text)

    def _reduce(self, qty, price):
        if self.is_open is False:
            raise EmptyPosition('The position is closed.')

        # just to prevent confusion
        qty = abs(qty)

        estimated_profit = jh.estimate_PNL(qty, self.entry_price, price,
                                           self.type)

        if self.exchange:
            self.exchange.increase_balance(
                self, qty * self.entry_price + estimated_profit)

        if self.type == trade_types.LONG:
            self.qty -= qty
        elif self.type == trade_types.SHORT:
            self.qty += qty

        info_text = 'REDUCED position: {}, {}, {}, {}, ${}'.format(
            self.exchange_name, self.symbol, self.type, self.qty, round(self.entry_price, 2)
        )

        if jh.is_debuggable('position_reduced'):
            logger.info(info_text)

        if jh.is_live() and config['env']['notifications']['events']['updated_position']:
            notifier.notify(info_text)

    def _increase(self, qty, price):
        if not self.is_open:
            raise OpenPositionError(
                'position must be already open in order to incrase its size')

        qty = abs(qty)
        size = qty * price

        if self.exchange:
            self.exchange.decrease_balance(self, size)

        self.entry_price = jh.estimate_average_price(qty, price, self.qty,
                                                     self.entry_price)

        if self.type == trade_types.LONG:
            self.qty += qty
        elif self.type == trade_types.SHORT:
            self.qty -= qty

        info_text = 'INCREASED position: {}, {}, {}, {}, ${}'.format(
            self.exchange_name, self.symbol, self.type, self.qty, round(self.entry_price, 2)
        )

        if jh.is_debuggable('position_increased'):
            logger.info(info_text)

        if jh.is_live() and config['env']['notifications']['events']['updated_position']:
            notifier.notify(info_text)

    def _open(self, qty, price, change_balance=True):
        if self.is_open:
            raise OpenPositionError('an already open position cannot be opened')

        if change_balance:
            size = abs(qty) * price
            if self.exchange:
                self.exchange.decrease_balance(self, size)

        self.entry_price = price
        self.exit_price = None
        self.qty = qty
        self.opened_at = jh.now()

        info_text = 'OPENED {} position: {}, {}, {}, ${}'.format(
            self.type, self.exchange_name, self.symbol, self.qty, round(self.entry_price, 2)
        )

        if jh.is_debuggable('position_opened'):
            logger.info(info_text)

        if jh.is_live() and config['env']['notifications']['events']['updated_position']:
            notifier.notify(info_text)

    def _on_opened_order(self, order):
        # skip MARKET orders because they're already
        # being impacted at self.on_executed_order
        if order.type == order_types.MARKET:
            return

        qty = order.qty
        price = order.price
        size = ju.qty_to_size(qty, price)

        # open-position order
        if self.qty == 0:
            if self.exchange:
                self.exchange.decrease_balance(self, size)
        # increase-position order
        elif self.qty * qty > 0:
            if self.exchange:
                self.exchange.decrease_balance(self, size)
        # reduce-position order
        elif self.qty * qty < 0:
            if abs(qty) > abs(self.qty):
                diff_qty = qty + self.qty
                size = ju.qty_to_size(diff_qty, price)
                if self.exchange:
                    self.exchange.decrease_balance(self, size)

    def _on_canceled_order(self, order):
        qty = order.qty
        price = order.price
        size = ju.qty_to_size(qty, price)

        if order.is_reduce_only:
            return

        # open-position order
        if self.qty == 0:
            if self.exchange:
                self.exchange.increase_balance(self, size, True)
        # increase-position order
        elif self.qty * qty > 0:
            if self.exchange:
                self.exchange.increase_balance(self, size, True)
        # reduce-position order
        elif self.qty * qty < 0:
            if abs(qty) > abs(self.qty):
                diff_qty = qty + self.qty
                size = ju.qty_to_size(diff_qty, price)
                if self.exchange:
                    self.exchange.increase_balance(self, size, True)

    def _on_executed_order(self, order: Order):
        qty = order.qty
        price = order.price

        # handle REDUCE_ONLY orders
        if order.flag == order_flags.REDUCE_ONLY:
            if self.qty == 0 or self.qty * qty > 0:
                return
            if self.qty * qty < 0 and abs(qty) > abs(self.qty):
                self._close(price)
                if self.strategy:
                    self.strategy._on_updated_position(order)
                return

        # order opens position
        if self.qty == 0 and order.type:
            change_balance = order.type == order_types.MARKET
            self._open(qty, price, change_balance)
        # order closes position
        elif (self.qty + qty) == 0:
            self._close(price)
        # order increases the size of the position
        elif self.qty * qty > 0:
            self._increase(qty, price)
        # order reduces the size of the position
        elif self.qty * qty < 0:
            # if size of the order is big enough to both close the
            # position AND open it on the opposite side
            if abs(qty) > abs(self.qty):
                diff_qty = qty + self.qty
                self._close(price)
                self._open(diff_qty, price)
            else:
                self._reduce(qty, price)

        if self.strategy:
            self.strategy._on_updated_position(order)



from collections import namedtuple

import numpy as np
import talib

from jesse.helpers import get_candle_source

KST = namedtuple('KST', ['line', 'signal'])


def kst(candles: np.ndarray, sma_period1=10, sma_period2=10, sma_period3=10, sma_period4=15, roc_period1=10, roc_period2=15, roc_period3=20, roc_period4=30, signal_period=9, source_type="close", sequential=False) -> KST:
    """
    Know Sure Thing (KST)

    :param candles: np.ndarray
    :param sma_period1: int - default=10
    :param sma_period2: int - default=10
    :param sma_period3: int - default=10
    :param sma_period4: int - default=15
    :param roc_period1: int - default=10
    :param roc_period2: int - default=15
    :param roc_period3: int - default=20
    :param roc_period4: int - default=30
    :param signal_period: int - default=9
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: KST(line, signal)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]


    source = get_candle_source(candles, source_type=source_type)


    aroc1 = talib.SMA(talib.ROC(source, timeperiod=roc_period1), sma_period1)
    aroc2 = talib.SMA(talib.ROC(source, timeperiod=roc_period2), sma_period2)
    aroc3 = talib.SMA(talib.ROC(source, timeperiod=roc_period3), sma_period3)
    aroc4 = talib.SMA(talib.ROC(source, timeperiod=roc_period4), sma_period4)
    line = aroc1[len(aroc1) - len(aroc4):] + 2 * aroc2[len(aroc2) - len(aroc4):] + \
           3 * aroc3[len(aroc3) - len(aroc4):] + 4 * aroc4
    signal = talib.SMA(line, signal_period)

    if sequential:
        return KST(line, signal)
    else:
        return KST(line[-1], signal[-1])
